"""The DDS response.

The DDS response builds a representation of the structure of the dataset,
informing which variables are contained, their shape, type and dimensions.
Together with the DAS, the DDS describes all metadata associated with a given
dataset, allowing clients to introspect the variables and request data as
necessary.

"""

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from ..lib import NUMPY_TO_DAP2_TYPEMAP, __version__
from ..model import BaseType, DatasetType, GridType, SequenceType, StructureType
from .lib import BaseResponse

INDENT = " " * 4


class DDSResponse(BaseResponse):
    """The DDS response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_dds"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in dds(self.dataset):
            yield line.encode("ascii")


@singledispatch
def dds(var, level=0, sequence=0):
    """Single dispatcher that generates the DDS response."""
    raise StopIteration


@dds.register(DatasetType)
def _(var, level=0, sequence=0):
    yield "{indent}Dataset {{\n".format(indent=level * INDENT)
    for child in var.children():
        for line in dds(child, level + 1, sequence):
            yield line
    yield "{indent}}} {name};\n".format(indent=level * INDENT, name=var.name)


@dds.register(SequenceType)
def _sequencetype(var, level=0, sequence=0):
    yield "{indent}Sequence {{\n".format(indent=level * INDENT)
    for child in var.children():
        for line in dds(child, level + 1, sequence + 1):
            yield line
    yield "{indent}}} {name};\n".format(indent=level * INDENT, name=var.name)


@dds.register(StructureType)
def _structuretype(var, level=0, sequence=0):
    yield "{indent}Structure {{\n".format(indent=level * INDENT)
    for child in var.children():
        for line in dds(child, level + 1, sequence):
            yield line
    yield "{indent}}} {name};\n".format(indent=level * INDENT, name=var.name)


@dds.register(GridType)
def _gridtype(var, level=0, sequence=0):
    yield "{indent}Grid {{\n".format(indent=level * INDENT)

    yield "{indent}Array:\n".format(indent=(level + 1) * INDENT)
    for line in dds(var.array, level + 2, sequence):
        yield line

    yield "{indent}Maps:\n".format(indent=(level + 1) * INDENT)
    for map_ in var.maps.values():
        for line in dds(map_, level + 2, sequence):
            yield line

    yield "{indent}}} {name};\n".format(indent=level * INDENT, name=var.name)


@dds.register(BaseType)
def _basetype(var, level=0, sequence=0):
    shape = var.shape[sequence:]

    if var.dimensions:
        shape = "".join(map("[{0[0]} = {0[1]}]".format, zip(var.dimensions, shape)))
    elif len(shape) == 1:
        shape = "[{0} = {1}]".format(var.name, shape[0])
    else:
        shape = "".join("[{0}]".format(len) for len in shape)

    yield "{indent}{type} {name}{shape};\n".format(
        indent=level * INDENT,
        type=NUMPY_TO_DAP2_TYPEMAP[var.dtype.char],
        name=var.name,
        shape=shape,
    )



"""Test the DAP handler, which forms the core of the client."""

import numpy as np
import pytest
from netCDF4 import Dataset

from pydap.handlers.netcdf import NetCDFHandler


@pytest.fixture(scope="module")
def simple_data():
    data = [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]
    return data


@pytest.fixture(scope="module")
def simple_nc_file(simple_data, tmpdir_factory):
    file_name = str(tmpdir_factory.mktemp("nc").join("simple.nc"))
    with Dataset(file_name, "w") as output:
        output.createDimension("index", None)
        temp = output.createVariable("index", "<i4", ("index",))
        split_data = zip(*simple_data)
        temp[:] = next(split_data)
        temp = output.createVariable("temperature", "<f8", ("index",))
        temp[:] = next(split_data)
        temp = output.createVariable("station", "S40", ("index",))
        for item_id, item in enumerate(next(split_data)):
            temp[item_id] = item
    return file_name


@pytest.fixture(scope="module")
def simple_Group_data():
    data = np.arange(10, 26, 1, dtype="f4").reshape(1, 4, 4)
    return data


@pytest.fixture(scope="module")
def simple_group_array_file(simple_Group_data, tmpdir_factory):
    file_name = str(tmpdir_factory.mktemp("nc").join("Group_array.nc"))
    with Dataset(file_name, "w") as output:
        output.createDimension("time", None)  # unlimited dimension
        output.createDimension("nv", 2)  # unlimited dimension
        output.createVariable("time", "<f8", ("time",))
        output.createVariable("time_nbds", "<f8", ("time", "nv"))
        group = output.createGroup("Group")
        group.createDimension("X", 4)
        group.createDimension("Y", 4)
        group.createVariable("X", "<i4", ("X",))
        group.createVariable("Y", "<i4", ("Y",))
        group.createVariable("temperature", np.float32, ("time", "Y", "X"))
        group["temperature"][:] = simple_Group_data
    return file_name


@pytest.fixture(scope="module")
def nested_group_array_file(simple_Group_data, tmpdir_factory):
    file_name = str(tmpdir_factory.mktemp("nc").join("NestedGroup_array.nc"))
    with Dataset(file_name, "w") as output:
        output.createDimension("time", None)  # unlimited dimension
        output.createVariable("time", "<f8", ("time",))
        group1 = output.createGroup("Group")
        subgroup1 = group1.createGroup("SubGroup")
        subgroup1.createDimension("X", 4)
        subgroup1.createDimension("Y", 4)
        subgroup1.createVariable("X", "<i4", ("X",))
        subgroup1.createVariable("Y", "<i4", ("Y",))
        subgroup1.createVariable("temperature", np.float32, ("time", "Y", "X"))
        subgroup1["temperature"][:] = simple_Group_data
    return file_name


@pytest.fixture(scope="module")
def nested_group_repeatdims_file(tmpdir_factory):
    file_name = str(tmpdir_factory.mktemp("nc").join("unaligned_subgroups.nc"))
    with Dataset(file_name, "w", format="NETCDF4") as root_group:
        group_1 = root_group.createGroup("/Group1")
        subgroup_1 = group_1.createGroup("/subgroup1")

        root_group.createDimension("lat", 1)
        root_group.createDimension("lon", 2)
        root_group.createVariable("root_variable", np.float64, ("lat", "lon"))

        group_1_var = group_1.createVariable("group_1_var", np.float64, ("lat", "lon"))
        group_1_var[:] = np.array([[0.1, 0.2]])
        group_1_var.units = "K"
        group_1_var.long_name = "air_temperature"

        subgroup_1.createDimension("lat", 2)

        subgroup1_var = subgroup_1.createVariable(
            "subgroup1_var", np.float64, ("lat", "lon")
        )
        subgroup1_var[:] = np.array([[0.1, 0.2]])
    return file_name


@pytest.fixture(scope="module")
def simple_handler(simple_nc_file):
    return NetCDFHandler(simple_nc_file)


@pytest.fixture(scope="module")
def simple_handler2(simple_group_array_file):
    return NetCDFHandler(simple_group_array_file)


@pytest.fixture(scope="module")
def simple_handler3(nested_group_array_file):
    return NetCDFHandler(nested_group_array_file)


@pytest.fixture(scope="module")
def simple_handler4(nested_group_repeatdims_file):
    return NetCDFHandler(nested_group_repeatdims_file)


def test_handler(simple_data, simple_handler):
    """Test that dataset has the correct data proxies for grids."""
    dataset = simple_handler.dataset
    dtype = [("index", "<i4"), ("temperature", "<f8"), ("station", "S40")]
    retrieved_data = list(
        zip(
            dataset["index"][:],
            dataset["temperature"][:],
            dataset["station"][:],
        )
    )
    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=dtype), np.array(simple_data, dtype=dtype)
    )


def test_handler_array(simple_Group_data, simple_handler2):
    """Test that dataset has the correct data proxies for grids."""
    dataset = simple_handler2.dataset
    r_dims = dataset.dimensions
    g_dims = dataset["Group"].attributes["dimensions"]
    temp = dataset["/Group/temperature"][:]
    np.testing.assert_array_equal(temp.data, simple_Group_data)
    assert r_dims == {"time": 1, "nv": 2}
    assert g_dims == {"X": 4, "Y": 4}
    assert temp.dimensions == ("/time", "/Group/Y", "/Group/X")


def test_handler_nested_Group_array(simple_Group_data, simple_handler3):
    """Test that dataset has the correct data proxies for grids."""
    dataset = simple_handler3.dataset
    r_dims = dataset.dimensions
    g_dims = dataset["Group"].attributes["dimensions"]
    sg_dims = dataset["/Group/SubGroup"].attributes["dimensions"]
    temp = dataset["/Group/SubGroup/temperature"][:]
    np.testing.assert_array_equal(temp.data, simple_Group_data)
    assert r_dims == {"time": 1}
    assert g_dims == {}
    assert sg_dims == {"X": 4, "Y": 4}
    assert temp.dimensions == ("/time", "/Group/SubGroup/Y", "/Group/SubGroup/X")


def test_handler_repeatdims_Group_array(simple_handler4):
    """Test that dataset has the correct data proxies for grids."""
    dataset = simple_handler4.dataset
    r_dims = dataset.dimensions
    g_dims = dataset["Group1"].dimensions
    var_g1 = dataset["Group1/group_1_var"].dimensions
    sg_dims = dataset["Group1/subgroup1"].dimensions
    var_sg1 = dataset["Group1/subgroup1/subgroup1_var"].dimensions
    assert r_dims == {"lat": 1, "lon": 2}
    assert g_dims == {}
    assert sg_dims == {"lat": 2}
    assert var_g1 == ("/lat", "/lon")
    assert var_sg1 == ("/Group1/subgroup1/lat", "/lon")


@pytest.fixture(scope="module")
def simple_application(simple_handler):
    from pydap.wsgi.ssf import ServerSideFunctions

    return ServerSideFunctions(simple_handler)


# def test_open(simple_data, simple_application):
#     """Test that NetCDFHandler can be read through open_url."""
#     dataset = DAPHandler("http://localhost:8001/", simple_application).dataset
#     dtype = [("index", "<i4"), ("temperature", "<f8"), ("station", "S40")]
#     retrieved_data = list(
#         zip(
#             dataset["index"][:],
#             dataset["temperature"][:],
#             dataset["station"][:],
#         )
#     )
#     np.testing.assert_array_equal(
#         np.array(retrieved_data, dtype=dtype), np.array(simple_data, dtype=dtype)
#     )



"""
This is an example of a Dapper-compliant dataset.

"""

import unittest

from pydap.parsers.das import add_attributes, parse_das
from pydap.parsers.dds import dds_to_dataset

DDS = """Dataset {
    Sequence {
        Float32 lat;
        Float64 time;
        Float32 lon;
        Int32 _id;
        Sequence {
            Float32 temp_qc;
            Float32 salinity_qc;
            Float32 oxygen;
            Float32 depth;
            Float32 temp;
            Float32 chlorophyll;
            Float32 salinity;
            Float32 oxygen_qc;
            Float32 pressure_qc;
            Float32 pressure;
            Float32 chlorophyll_qc;
        } profile;
        Structure {
            String title;
            String data_origin;
            String source;
            String Conventions;
            String country_code;
            String cruise;
            String cruise_number;
            String cast_number;
        } attributes;
        Structure {
            Structure {
                Float32 valid_range[2];
            } depth;
        } variable_attributes;
    } location;
    Structure {
        Float32 lon_range[2];
        Float32 lat_range[2];
        Float32 depth_range[2];
        Float64 time_range[2];
    } constrained_ranges;
} northPacific;"""

DAS = """Attributes {
    NC_GLOBAL {
        Int32 max_profiles_per_request 15000;
        Int32 total_profiles_in_dataset 1707616;
        String version "1.1.0";
        String owner "";
        String contact "";
        String Conventions "epic-insitu-1.0";
        Float64 lon_range 100.0, 289.180000305176;
        Float64 lat_range 1.20000004244503E-5, 90.0;
        Float64 depth_range 0.0, 10860.900390625;
        Float64 time_range -1.1432412E12, 1.10778048E12;
    }
    location._id {
        String long_name "sequence id";
        Int32 missing_value 2147483647;
        String units "";
    }
    location.lat {
        String units "degree_north";
        String long_name "latitude";
        Float32 missing_value NaN;
        String axis "Y";
    }
    location.profile.temp_qc {
        String units "";
        String long_name "temperature quality flag";
        Float32 missing_value NaN;
    }
    location.profile.salinity_qc {
        String units "";
        String long_name "salinity quality flag";
        Float32 missing_value NaN;
    }
    location.time {
        String units "msec since 1970-01-01 00:00:00 GMT";
        String long_name "time";
        Float64 missing_value NaN;
        String axis "T";
    }
    location.profile.oxygen {
        String units "ml/l";
        String long_name "oxygen";
        Float32 missing_value NaN;
    }
    location.profile.depth {
        String units "m";
        String long_name "depth";
        Float32 missing_value NaN;
        String axis "Z";
    }
    location.lon {
        String units "degree_east";
        String long_name "longitude";
        Float32 missing_value NaN;
        String axis "X";
    }
    location.profile.temp {
        String units "degc";
        String long_name "temperature";
        Float32 missing_value NaN;
    }
    location.profile.chlorophyll {
        String units "ug/l";
        String long_name "total chlorophyll";
        Float32 missing_value NaN;
    }
    location.profile.salinity {
        String units "psu";
        String long_name "salinity";
        Float32 missing_value NaN;
    }
    location.profile.oxygen_qc {
        String units "";
        String long_name "oxygen quality flag";
        Float32 missing_value NaN;
    }
    location.profile.pressure_qc {
        String units "";
        String long_name "pressure quality flag";
        Float32 missing_value NaN;
    }
    location.profile.pressure {
        String units "dbar";
        String long_name "pressure";
        Float32 missing_value NaN;
    }
    location.profile.chlorophyll_qc {
        String units "";
        String long_name "total chlorophyll quality flag";
        Float32 missing_value NaN;
    }
}"""


class TestDapper(unittest.TestCase):
    def setUp(self):
        dataset = dds_to_dataset(DDS)
        attributes = parse_das(DAS)
        self.dataset = add_attributes(dataset, attributes)

    def test_parse(self):
        self.assertEqual(list(self.dataset.keys()), ["location", "constrained_ranges"])



import math
import multiprocessing
import socket
import sys
import threading
import time
from wsgiref.simple_server import make_server

import numpy as np
from webob.exc import HTTPError
from webob.request import Request

from ..handlers.lib import BaseHandler
from ..model import BaseType, DatasetType
from ..net import get_response

DefaultDataset = DatasetType("Default")
DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))


def get_open_port():
    # http://stackoverflow.com/questions/2838244/
    # get-open-tcp-port-in-python/2838309#2838309
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def run_server_in_process(httpd, shutdown, **kwargs):
    _server = threading.Thread(target=httpd.serve_forever, kwargs=kwargs)
    _server.start()
    shutdown.wait()
    httpd.shutdown()
    _server.join()


class LocalTestServer(object):
    """
    Simple server instance that can be used to test pydap.
    Relies on threading and is usually slow (it has to
    start and shutdown which typically takes ~2 sec).

    Usage:
    >>> import numpy as np
    >>> from pydap.handlers.lib import BaseHandler
    >>> from pydap.model import DatasetType, BaseType
    >>> DefaultDataset = DatasetType("Default")
    >>> DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
    >>> DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
    >>> DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))
    >>> DefaultDataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> application = BaseHandler(DefaultDataset)
    >>> from pydap.client import open_url

    As an instance:
    >>> with LocalTestServer(application) as server:
    ...     dataset = open_url("http://localhost:%s" % server.port)
    ...     dataset
    ...     print(dataset['byte'].data[:])
    ...     print(dataset['string'].data[:])
    ...     print(dataset['short'].data[:])
    <DatasetType with children 'byte', 'string', 'short'>
    [0 1 2 3 4]
    [b'one' b'two']
    1

    Or by managing connection and deconnection:
    >>> server = LocalTestServer(application)
    >>> server.start()
    >>> dataset = open_url("http://localhost:%s" % server.port)
    >>> dataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> print(dataset['byte'].data[:])
    [0 1 2 3 4]
    >>> server.shutdown()
    """

    def __init__(
        self,
        application=BaseHandler(DefaultDataset),
        port=None,
        wait=0.5,
        polling=1e-2,
        as_process=False,
    ):
        self._port = port or get_open_port()
        self.application = application
        self._wait = wait
        self._polling = polling
        self._as_process = as_process
        self._address = "0.0.0.0"

    @property
    def url(self):
        return "http://{0}:{1}/".format(self._address, self.port)

    def start(self):
        # Start a simple WSGI server:
        application = self.application

        self._httpd = make_server(self._address, self.port, application)
        kwargs = {"poll_interval": 0.1}

        if self._as_process:
            self._shutdown = multiprocessing.Event()
            if sys.platform in ["darwin", "win32"]:
                # see https://github.com/python/cpython/issues/77906
                # no long term solution, simply temporaty fix
                ctx = multiprocessing.get_context("fork")
                Process = ctx.Process
            else:
                Process = multiprocessing.Process
            self._server = Process(
                target=run_server_in_process,
                args=(self._httpd, self._shutdown),
                kwargs=kwargs,
            )
        else:
            self._server = threading.Thread(
                target=self._httpd.serve_forever, kwargs=kwargs
            )

        self._server.start()
        self.poll_server()

    def poll_server(self):
        # Poll the server
        ok = False
        for trial in range(int(math.ceil(self._wait / self._polling))):
            try:
                # When checking whether server has started, do
                # not verify ssl:
                resp = get_response(
                    Request.blank(self.url + ".dds"), None, verify=False
                )
                ok = resp.status_code == 200
            except HTTPError:
                pass
            if ok:
                break
            time.sleep(self._polling)

        if not ok:
            self.shutdown()
            raise Exception(
                (
                    "LocalTestServer did not start in {0}s. "
                    "Try using LocalTestServer(..., wait={1}"
                ).format(self._wait, 2 * self._wait)
            )

    @property
    def port(self):
        return self._port

    def __enter__(self):
        self.start()
        return self

    def shutdown(self):
        # Tell the server to shutdown:
        if self._as_process:
            self._shutdown.set()
        else:
            self._httpd.shutdown()
        self._server.join()
        self._httpd.server_close()

    def __exit__(self, *_):
        self.shutdown()



import re
import sys
import unittest

from webob import Request

from pydap.lib import __version__
from pydap.responses.error import ErrorResponse


class TestErrorResponse(unittest.TestCase):
    def setUp(self):
        # create an exception that would happen in runtime
        try:
            1 / 0
        except Exception:
            error = ErrorResponse(sys.exc_info())

        req = Request.blank('/')
        self.res = req.get_response(error)

    def test_status(self):
        self.assertEqual(self.res.status, "500 Internal Error")

    def test_content_type(self):
        self.assertEqual(self.res.content_type, "text/plain")

    def test_charset(self):
        self.assertEqual(self.res.charset, "utf-8")

    def test_headers(self):
        self.assertEqual(self.res.headers['Content-Type'],
                         'text/plain; charset=utf-8')
        self.assertEqual(self.res.headers['Content-description'], 'OPeNDAP_error')
        self.assertEqual(self.res.headers['OPeNDAP-Server'],
                         'pydap/' + __version__)

    def test_body(self):
        self.assertRegex(self.res.text, re.compile(r"""Error {
    code = -1;
    message = "Traceback \(most recent call last\):
  File .*
    1 \/ 0
(    \~\~\^\~\~
)?ZeroDivisionError:( integer)? division( or modulo)? by zero
";
}"""), re.DOTALL)


if __name__ == '__main__':
    pass



"""Response with information about pydap version."""

import sys
from importlib.metadata import entry_points
from json import dumps

from webob import Response

from ..lib import __dap__, __version__


class VersionResponse(object):
    """A specialized response for debugging.

    This is a special response used to display information about the server.

    """

    __version__ = __version__

    def __init__(self, dataset):
        output = {
            "pydap": __version__,
            "dap": __dap__,
            "handlers": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in entry_points(group="pydap.handler")
            ),
            "responses": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in entry_points(group="pydap.response")
            ),
            "functions": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in entry_points(group="pydap.function")
            ),
            "python": sys.version,
        }
        self.body = dumps(output, indent=4)

    def __call__(self, environ, start_response):
        res = Response()
        res.text = str(self.body)
        res.status = "200 OK"
        res.content_type = "application/json"
        res.charset = "utf-8"
        res.headers.add("Content-description", "dods_version")
        res.headers.add("OPeNDAP-Server", "pydap/%s" % __version__)

        return res(environ, start_response)



from . import get_cookies


def setup_session(username, password, check_url=None, session=None, verify=True):
    """
    A special call to get_cookies.setup_session that is tailored for
    URS EARTHDATA at NASA credentials.

    Parameters:
    ----------
    username: str
    password: str
    check_url: str, None (default)
    session: `requests.session`, None (default)
    verify: bool, True (default)

    Example Data Access and Authentification in:
    see `https://opendap.github.io/documentation/tutorials/`

    """

    if session is not None:
        # URS connections cannot be kept alive at the moment.
        session.headers.update({"Connection": "close"})
    session = get_cookies.setup_session(
        "https://urs.earthdata.nasa.gov",
        username=username,
        password=password,
        session=session,
        check_url=check_url,
        verify=verify,
    )
    return session



import codecs
import pickle


def main():
    logfile = "dap4_access_test.log"

    with open(logfile, "r") as f:
        f = f.read()

    p = f.split("----BEGIN PICKLE-----")[1].split("----END PICKLE-----")[0]
    p = "".join(p).strip().replace("\n", "")

    response = pickle.loads(codecs.decode(p.encode("utf-8"), "base64"))

    print(dict(response.headers))
    with open("out.html", "w") as f:
        f.write(response.text)

    with open("response.pickle", "wb") as f:
        pickle.dump(response, f)


if __name__ == "__main__":
    main()



import unittest

import numpy as np
from webob.request import Request

from pydap.client import open_url
from pydap.handlers.lib import BaseHandler
from pydap.model import BaseType, DatasetType


class TestQuote(unittest.TestCase):
    def setUp(self):
        # create dataset
        self.dataset = DatasetType("test")
        self.dataset["foo["] = BaseType("foo[", np.array(1))

        # create WSGI app
        self.app = BaseHandler(self.dataset)

    def test_name(self):
        """Check that the name was properly escaped."""
        self.assertEqual(list(self.dataset.keys()), ["foo%5B"])

    def test_dds(self):
        text = Request.blank("/.dds").get_response(self.app).text
        self.assertEqual(
            text,
            """Dataset {
    Int32 foo%5B;
} test;
""",
        )

    def test_request(self):
        text = Request.blank("/.dds?foo%5B").get_response(self.app).text
        self.assertEqual(
            text,
            """Dataset {
    Int32 foo%5B;
} test;
""",
        )

    def test_client(self):
        dataset = open_url("http://localhost:8001/", application=self.app)
        self.assertEqual(list(self.dataset.keys()), ["foo%5B"])
        self.assertEqual(dataset["foo["].name, "foo%5B")
        self.assertEqual(dataset["foo%5B"].data, 1)


class TestPeriod(unittest.TestCase):
    def setUp(self):
        # create dataset
        dataset = DatasetType("test")
        dataset["a.b"] = BaseType("a.b", np.array(1))

        # create WSGI app
        self.app = BaseHandler(dataset)

    def test_dds(self):
        text = Request.blank("/.dds").get_response(self.app).text
        self.assertEqual(
            text,
            """Dataset {
    Int32 a%2Eb;
} test;
""",
        )

    def test_client(self):
        dataset = open_url("http://localhost:8001/", application=self.app)
        self.assertEqual(dataset["a.b"].name, "a%2Eb")
        # the test below gives problem because a.b is passed
        # to the `query` part of the constrain. However, this a.b
        # in the query conflicts with the behavior of a being a sequence
        # and b being a variable within the sequence...
        # self.assertEqual(dataset["a.b"][0], 1)



"""Test basic handler functions."""

import copy
import unittest

import numpy as np
from webtest import AppError
from webtest import TestApp as App

from pydap.exceptions import ConstraintExpressionError
from pydap.handlers.lib import (
    BaseHandler,
    ConstraintExpression,
    IterData,
    apply_projection,
    apply_selection,
)
from pydap.lib import walk
from pydap.model import BaseType, SequenceType, StructureType
from pydap.parsers import parse_projection
from pydap.tests.datasets import (
    NestedSequence,
    SimpleArray,
    SimpleGrid,
    SimpleSequence,
    SimpleStructure,
    VerySimpleSequence,
)


class TestBaseHandler(unittest.TestCase):
    """Test the base handler as a WSGI app."""

    def setUp(self):
        """Create a basic WSGI app."""
        self.app = App(MockHandler(SimpleArray))

    def test_unconstrained_das(self):
        """DAS responses are always unconstrained."""
        res = self.app.get("/.dds")
        self.assertEqual(
            res.text,
            """Dataset {
    Byte byte[byte = 5];
    String string[string = 2];
    Int16 short;
} SimpleArray;
""",
        )

        res = self.app.get("/.dds?byte")
        self.assertEqual(
            res.text,
            """Dataset {
    Byte byte[byte = 5];
} SimpleArray;
""",
        )

        res = self.app.get("/.das")
        das = res.text
        self.assertEqual(
            das,
            """Attributes {
    byte {
    }
    string {
    }
    short {
    }
}
""",
        )

        # check that DAS is unmodifed with constraint expression
        res = self.app.get("/.das?byte")
        self.assertEqual(res.text, das)

    def test_exception(self):
        """
        Test exception handling.

        By default pydap will capture all exceptions and return a formatted
        error response.

        """
        with self.assertRaises(AppError):
            self.app.get("/.foo")

    def test_exception_non_captured(self):
        """Test exception handling when not captured."""
        app = App(
            MockHandler(SimpleArray), extra_environ={"x-wsgiorg.throw_errors": True}
        )
        with self.assertRaises(KeyError):
            app.get("/.foo")

    def test_missing_dataset(self):
        """Test exception when dataset is not set."""
        app = App(MockHandler(), extra_environ={"x-wsgiorg.throw_errors": True})
        with self.assertRaises(NotImplementedError):
            app.get("/.dds")


class TestApplySelection(unittest.TestCase):
    """Test function that applies selections to the dataset."""

    def setUp(self):
        """Build our own sequence.

        pydap uses lightweight copies of objects that share data. This breaks
        unit tests since the same objects are reused for tests.

        """
        # make a dataset with its own data
        self.dataset = copy.copy(SimpleSequence)
        self.dataset.cast.data = SimpleSequence.cast.data.copy()

    def test_no_selection(self):
        """Test no selection in the query string."""
        dataset = apply_selection("", self.dataset)
        np.testing.assert_array_equal(dataset.cast.data, self.dataset.cast.data)

    def test_simple_selection(self):
        """Test a simple selection applied to the dataset."""
        dataset = apply_selection(["cast.lon=100"], self.dataset)
        np.testing.assert_array_equal(
            dataset.cast.data,
            self.dataset.cast.data[self.dataset.cast.data["lon"] == 100],
        )

    def test_multiple_selections(self):
        """Test multiple selections applied to dataset."""
        dataset = apply_selection(["cast.lon=100", "cast.lat>0"], self.dataset)
        np.testing.assert_array_equal(
            dataset.cast.data,
            self.dataset.cast.data[self.dataset.cast.data["lon"] == 100][
                self.dataset.cast.data["lat"] > 0
            ],
        )


class TestApplyProjectionGrid(unittest.TestCase):
    """Test applying projections on a dataset with a grid."""

    def setUp(self):
        """Build dataset with no shared data."""
        self.dataset = copy.copy(SimpleGrid)
        for var in walk(self.dataset, BaseType):
            var.data = var.data.copy()

    def test_no_projection(self):
        """Test no projections."""
        dataset = apply_projection("", self.dataset)
        self.assertEqual(list(dataset.children()), [])

    def test_simple_projection(self):
        """Test simple projections."""
        dataset = apply_projection(parse_projection("x"), self.dataset)
        self.assertEqual(list(dataset.keys()), ["x"])

    def test_simple_projection_with_index(self):
        """Test simple projections."""
        dataset = apply_projection(parse_projection("x[1]"), self.dataset)
        np.testing.assert_array_equal(dataset.x.data, [1])

    def test_array(self):
        """Test that the grid degenerates into a structure."""
        dataset = apply_projection(
            parse_projection("SimpleGrid.SimpleGrid"), self.dataset
        )
        self.assertIsInstance(dataset.SimpleGrid, StructureType)

    def test_array_slice(self):
        """Test slices applied to a grid."""
        dataset = apply_projection(parse_projection("SimpleGrid[1]"), self.dataset)
        np.testing.assert_array_equal(
            dataset.SimpleGrid.x.data, self.dataset.SimpleGrid[1].x.data
        )
        np.testing.assert_array_equal(
            dataset.SimpleGrid.y.data, self.dataset.SimpleGrid[1].y.data
        )
        np.testing.assert_array_equal(
            dataset.SimpleGrid.SimpleGrid.data,
            self.dataset.SimpleGrid[1:2].SimpleGrid.data,
        )


class TestApplyProjectionSequence(unittest.TestCase):
    """Test applying projections on a dataset with a sequence."""

    def setUp(self):
        """Build dataset with no shared data."""
        self.dataset = copy.copy(VerySimpleSequence)
        self.dataset.sequence.data = VerySimpleSequence.sequence.data.copy()

    def test_sequence_projection(self):
        """Test projection slicing on sequences."""
        dataset = apply_projection(parse_projection("sequence[2]"), self.dataset)
        np.testing.assert_array_equal(
            dataset.sequence.data, VerySimpleSequence.sequence.data[2]
        )


class TestInvalidProjection(unittest.TestCase):
    """Test applying a projection to a structure object."""

    def test_structure_projection(self):
        """Test projection slicing on a structure."""
        with self.assertRaises(ConstraintExpressionError):
            apply_projection(parse_projection("types[0]"), SimpleStructure)


class TestConstraintExpression(unittest.TestCase):
    """Test the constraint expression object."""

    def test_str(self):
        """Test string representation."""
        ce = ConstraintExpression("a>1")
        self.assertEqual(str(ce), "a>1")

    def test_unicode(self):
        """Test unicode representation."""
        ce = ConstraintExpression("a>1")
        self.assertEqual(str(ce), "a>1")

    def test_and(self):
        """Test CE addition."""
        ce1 = ConstraintExpression("a>1")
        ce2 = ConstraintExpression("b>0")
        ce3 = ce1 & ce2
        self.assertEqual(str(ce3), "a>1&b>0")

    def test_or(self):
        """Expressions cannot be ORed."""
        ce1 = ConstraintExpression("a>1")
        ce2 = ConstraintExpression("b>0")
        with self.assertRaises(ConstraintExpressionError):
            ce1 | ce2


class TestIterData(unittest.TestCase):
    """
    Test the ``IterData`` class, used to store flat/nested sequence data.

    A flat ``IterData`` should behave like a Numpy structured array, except
    all operations are stored to be lazily evaluated when the object is
    iterated over.

    """

    def setUp(self):
        """Create a flat IterData."""
        template = SequenceType("a")
        template["b"] = BaseType("b")
        template["c"] = BaseType("c")
        template["d"] = BaseType("d")
        self.data = IterData([(1, 2, 3), (4, 5, 6)], template)

        self.array = np.array(
            np.rec.fromrecords(
                [
                    (1, 2, 3),
                    (4, 5, 6),
                ],
                names=["b", "c", "d"],
            )
        )

    def assertIteratorEqual(self, it1, it2):
        self.assertEqual(list(it1), list(it2))

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(repr(self.data), "<IterData to stream [(1, 2, 3), (4, 5, 6)]>")

    def test_dtype(self):
        """Test the ``dtype`` property."""
        self.assertEqual(self.data["b"].dtype, self.array["b"].dtype)

    def test_iteration(self):
        """Test iteration over data."""
        self.assertIteratorEqual(map(tuple, self.data), map(tuple, self.array))
        self.assertIteratorEqual(self.data["b"], self.array["b"])

    def test_filter(self):
        """Test filtering the object."""
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] == 1]),
            map(tuple, self.array[self.array["b"] == 1]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] != 1]),
            map(tuple, self.array[self.array["b"] != 1]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] >= 1]),
            map(tuple, self.array[self.array["b"] >= 1]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] <= 1]),
            map(tuple, self.array[self.array["b"] <= 1]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] > 1]),
            map(tuple, self.array[self.array["b"] > 1]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] < 1]),
            map(tuple, self.array[self.array["b"] < 1]),
        )

    def test_slice(self):
        """Test slicing the object."""
        self.assertIteratorEqual(map(tuple, self.data[1:]), map(tuple, self.array[1:]))

    def test_integer_slice(self):
        """Test slicing with an integer.

        Note that the behavior here is different from Numpy arrays, since the
        data access to ``IterData`` is through iteration it has no direct index
        access.

        """
        self.assertIteratorEqual(self.data[0:1], self.array[0:1].tolist())
        self.assertIteratorEqual(self.data[0], self.array[0:1].tolist())

    def test_invalid_child(self):
        """Test accessing a non-existing child."""
        with self.assertRaises(KeyError):
            self.data["e"]

    def test_invalid_key(self):
        """Test accessing using an invalid key."""
        with self.assertRaises(KeyError):
            self.data[(1, 2)]

    def test_selecting_children(self):
        """Test that we can select children."""
        self.assertIteratorEqual(
            map(tuple, self.data[["d", "b"]]), map(tuple, self.array[["d", "b"]])
        )

    def test_invalid_selection(self):
        """Test invalid selections.

        In theory this should never happen, since ``ConstraintExpression``
        object are constructly directly from existing children.

        """
        with self.assertRaises(ConstraintExpressionError):
            self.data[ConstraintExpression("a.e<1")]
        with self.assertRaises(ConstraintExpressionError):
            self.data[ConstraintExpression("a.d<foo")]

    def test_intercomparison_selection(self):
        """Test comparing children in the selection."""
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] == self.data["c"]]),
            map(tuple, self.array[self.array["b"] == self.array["c"]]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] != self.data["c"]]),
            map(tuple, self.array[self.array["b"] != self.array["c"]]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] >= self.data["c"]]),
            map(tuple, self.array[self.array["b"] >= self.array["c"]]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] <= self.data["c"]]),
            map(tuple, self.array[self.array["b"] <= self.array["c"]]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] > self.data["c"]]),
            map(tuple, self.array[self.array["b"] > self.array["c"]]),
        )
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] < self.data["c"]]),
            map(tuple, self.array[self.array["b"] < self.array["c"]]),
        )

    def test_selection_not_in_projection(self):
        """Test selection with variables that are not in the projection."""
        self.data[["d", "b"]]
        filtered = self.data[["d", "b"]][self.data["c"] > 3]
        self.assertEqual(filtered, self.array[["d", "b"]][self.array["c"] > 3])


class TestRegexp(unittest.TestCase):
    """Test regular expression match."""

    def test_regexp(self):
        sequence = SequenceType("sequence")
        sequence["name"] = BaseType("name")
        sequence.data = IterData(
            [
                ("John", "Paul", "George", "Ringo"),
            ],
            sequence,
        )

        filtered = sequence[ConstraintExpression('sequence.name=~"J.*"')]
        self.assertEqual(list(filtered.iterdata()), [("John",)])


class TestNestedIterData(unittest.TestCase):
    """Test ``IterData`` with nested data."""

    def setUp(self):
        """Load data from test dataset."""
        self.data = NestedSequence.location.data

    def test_iteration(self):
        """Test basic iteration."""
        self.assertEqual(
            list(self.data),
            [
                (1, 1, 1, [(10, 11, 12), (21, 22, 23)]),
                (2, 4, 4, [(15, 16, 17)]),
                (3, 6, 9, []),
                (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
            ],
        )

    def test_children_data(self):
        """Test getting data from a simple child."""
        self.assertEqual(list(self.data["lat"]), [1, 2, 3, 4])

    def test_sequence_children_data(self):
        """Test getting data from a sequence child."""
        self.assertEqual(
            list(self.data["time_series"]),
            [
                [(10, 11, 12), (21, 22, 23)],
                [(15, 16, 17)],
                [],
                [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)],
            ],
        )

    def test_deep_children_data(self):
        """Test getting data from a sequence child."""
        self.assertEqual(
            list(self.data["time_series"]["time"]),
            [[10, 21], [15], [], [31, 41, 51, 61]],
        )

    def test_selecting_children(self):
        """Test that we can select children."""
        self.assertEqual(
            list(self.data[["time_series", "elev"]]),
            [
                ([(10, 11, 12), (21, 22, 23)], 1),
                ([(15, 16, 17)], 4),
                ([], 9),
                ([(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)], 16),
            ],
        )

    def test_slice(self):
        """Test slicing the object."""
        self.assertEqual(
            list(self.data[1::2]),
            [
                (2, 4, 4, [(15, 16, 17)]),
                (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
            ],
        )

    def test_children_data_from_slice(self):
        """Test getting children data from a sliced sequence."""
        self.assertEqual(list(self.data[1::2]["lat"]), [2, 4])

    def test_sequence_children_data_from_slice(self):
        """Test getting children data from a sliced sequence."""
        self.assertEqual(
            list(self.data[1::2]["time_series"]),
            [[(15, 16, 17)], [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]],
        )

    def test_deep_slice(self):
        """Test slicing the inner sequence."""
        self.assertEqual(
            list(self.data["time_series"][::2]), [[(10, 11, 12), (21, 22, 23)], []]
        )

    def test_integer_slice(self):
        """Test slicing with an integer."""
        self.assertEqual(list(self.data["time_series"][1]), [[(15, 16, 17)]])

    def test_filter_data(self):
        """Test filtering the data."""
        self.assertEqual(
            list(self.data[self.data["lat"] > 2]),
            [
                (3, 6, 9, []),
                (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
            ],
        )

    def test_deep_filter(self):
        """Test deep filtering the data."""
        self.assertEqual(
            list(self.data[self.data["time_series"]["slp"] > 11]),
            [
                (1, 1, 1, [(21, 22, 23)]),
                (2, 4, 4, [(15, 16, 17)]),
                (3, 6, 9, []),
                (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
            ],
        )


class MockWorkingSet(object):
    """A fake working set for testing handlers."""

    def iter_entry_points(self, group):
        """Return a mock entry point."""
        yield MockEntryPoint(MockHandler)


class MockHandler(BaseHandler):
    """A fake handler for testing."""

    extensions = r"^.*\.foo$"

    def __init__(self, dataset=None):
        BaseHandler.__init__(self, dataset)
        self.additional_headers = [("X-debug", "True")]


class MockEntryPoint(object):
    """A fake entry point for testing."""

    def __init__(self, handler):
        self.handler = handler
        self.name = "test"
        self.module_name = "pydap.handlers.test"
        self.attrs = ("TestHandler",)

    def load(self):
        """Return the wrapped handler."""
        return self.handler



"""A parser for the Dataset Attribute Structure (DAS) response.

This module implements a DAS parser. The ``parse_das`` function will convert a
DAS response into a dictionary of attributes, which can be applied to an
existing dataset using the ``add_attributes`` function.

"""

import ast
import operator
import re
from functools import reduce

from ..lib import walk
from . import SimpleParser


class DASParser(SimpleParser):
    """A parser for the Dataset Attribute Structure response."""

    def __init__(self, das):
        super(DASParser, self).__init__(das, re.IGNORECASE | re.VERBOSE | re.DOTALL)

    def consume(self, regexp):
        """Return a token from the buffer.

        Not that it will Ignore white space when consuming tokens.

        """
        token = super(DASParser, self).consume(regexp)
        self.buffer = self.buffer.lstrip()
        return token

    def parse(self):
        """Start the parsing, returning a nested dictionary of attributes."""
        out = {}
        self.consume("attributes")
        self.container(out)
        return out

    def container(self, target):
        """Collect the attributes for a DAP variable."""
        self.consume("{")
        while not self.peek("}"):
            if self.peek(r"[^\s]+\s+{"):
                name = self.consume(r"[^\s]+")
                target[name] = {}
                self.container(target[name])
            else:
                name, values = self.attribute()
                target[name] = values
        self.consume("}")

    def attribute(self):
        """Parse attributes.

        The function will parse attributes from the DAS, converting them to the
        corresponding Python object. Returns the name of the attribute and the
        attribute(s).

        """
        type = self.consume(r"[^\s]+")
        name = self.consume(r"[^\s]+")

        values = []
        while not self.peek(";"):
            value = self.consume(
                r"""
                    ""          # empty attribute
                    |           # or
                    ".*?[^\\]"  # from quote up to an unquoted quote
                    |           # or
                    [^;,]+      # up to semicolon or comma
                """
            )

            if type.lower() in ["string", "url"]:
                value = str(value).strip('"')
            elif value.lower() in ["nan", "nan.", "-nan"]:
                value = float("nan")
            elif value.lower() in ["inf", "inf."]:
                value = float("inf")
            elif value.lower() in ["-inf", "-inf."]:
                value = float("-inf")
            else:
                value = ast.literal_eval(value)

            values.append(value)
            if self.peek(","):
                self.consume(",")

        self.consume(";")

        if len(values) == 1:
            values = values[0]

        return name, values


def parse_das(das):
    """Parse the DAS, returning nested dictionaries."""
    return DASParser(das).parse()


def add_attributes(dataset, attributes):
    """Add attributes from a parsed DAS to a dataset.

    Returns the dataset with added attributes.

    """
    dataset.attributes["NC_GLOBAL"] = attributes.get("NC_GLOBAL", {})
    dataset.attributes["DODS_EXTRA"] = attributes.get("DODS_EXTRA", {})

    for var in list(walk(dataset))[::-1]:
        # attributes can be flat, eg, "foo.bar" : {...}
        if var.id in attributes:
            var.attributes.update(attributes.pop(var.id))

        # or nested, eg, "foo" : { "bar" : {...} }
        try:
            nested = reduce(operator.getitem, [attributes] + var.id.split(".")[:-1])
            k = var.id.split(".")[-1]
            value = nested.pop(k)
        except KeyError:
            pass
        else:
            try:
                var.attributes.update(value)
            except (TypeError, ValueError):
                # This attribute should be given to the parent.
                # Keep around:
                nested.update({k: value})

    # add attributes that don't belong to any child
    for k, v in attributes.items():
        dataset.attributes[k] = v

    return dataset



"""Test ASCII response."""

import sys
import unittest

from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__
from pydap.responses.ascii import ascii
from pydap.tests.datasets import SimpleGrid, SimpleSequence


class TestASCIIResponseSequence(unittest.TestCase):
    """Test ASCII response from a sequence."""

    def setUp(self):
        """Create a simple WSGI app for testing."""
        app = App(BaseHandler(SimpleSequence))
        self.res = app.get("/.asc")

    def test_dispatcher(self):
        """Test the single dispatcher."""
        with self.assertRaises(StopIteration):
            ascii(None)

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "text/plain")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, "ascii")

    def test_headers(self):
        """Test headers from the response."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "dods_ascii"),
                    ("Content-type", "text/plain; charset=ascii"),
                    ("Content-Length", "440"),
                ]
            ),
        )

    def test_body(self):
        """Test the generated ASCII response."""
        self.assertEqual(
            self.res.text,
            """Dataset {
    Sequence {
        String id;
        Int32 lon;
        Int32 lat;
        Int32 depth;
        Int32 time;
        Int32 temperature;
        Int32 salinity;
        Int32 pressure;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.id, cast.lon, cast.lat, cast.depth, cast.time, cast.temperature, cast.salinity, cast.pressure
"1", 100, -10, 0, -1, 21, 35, 0
"2", 200, 10, 500, 1, 15, 35, 100

""",
        )


class TestASCIIResponseGrid(unittest.TestCase):
    """Test ASCII response from a grid."""

    def test_body(self):
        """Test the generated ASCII response."""
        app = App(BaseHandler(SimpleGrid))
        res = app.get("/.asc")
        self.assertEqual(
            res.text,
            """Dataset {
    Grid {
        Array:
            Int32 SimpleGrid[y = 2][x = 3];
        Maps:
            Int32 x[x = 3];
            Int32 y[y = 2];
    } SimpleGrid;
    Int32 x[x = 3];
    Int32 y[y = 2];
} SimpleGrid;
---------------------------------------------
SimpleGrid.SimpleGrid
[0][0] 0
[0][1] 1
[0][2] 2
[1][0] 3
[1][1] 4
[1][2] 5

SimpleGrid.x
[0] 0
[1] 1
[2] 2

SimpleGrid.y
[0] 0
[1] 1


x
[0] 0
[1] 1
[2] 2

y
[0] 0
[1] 1

""",
        )



"""HTML DAP response

This is a simple HTML response that allows that to be analysed on the browser.
The user can select a subset of the data and download in different formats.

"""

from jinja2 import ChoiceLoader, Environment, PackageLoader
from requests.utils import unquote
from webob import Response
from webob.dec import wsgify
from webob.exc import HTTPSeeOther

from ...lib import __version__
from ..lib import BaseResponse


class HTMLResponse(BaseResponse):
    """A simple HTML response for browsing and downloading data."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "DAP_form"),
                ("Content-type", "text/html; charset=utf-8"),
            ]
        )

        # our default environment; we need to include the base template from
        # pydap as well since our template extends it
        self.loaders = [
            PackageLoader("pydap.responses.html", "templates"),
            PackageLoader("pydap.wsgi", "templates"),
        ]

    @wsgify
    def __call__(self, req):
        # if request is a post we should redict to ASCII response
        if req.method == "POST":
            return self.redirect(req)

        # check if the server has specified a render environment; if it has,
        # make a copy and add our loaders to it
        if "pydap.jinja2.environment" in req.environ:
            env = req.environ["pydap.jinja2.environment"].overlay()
            env.loader = ChoiceLoader(
                [loader for loader in [env.loader] + self.loaders if loader]
            )
        else:
            env = Environment(loader=ChoiceLoader(self.loaders))

        env.filters["unquote"] = unquote
        template = env.get_template("html.html")

        tokens = req.path_info.split("/")[1:]
        breadcrumbs = [
            {
                "url": "/".join([req.application_url] + tokens[: i + 1]),
                "title": token,
            }
            for i, token in enumerate(tokens)
            if token
        ]

        context = {
            "root": req.application_url,
            "location": req.path_url,
            "breadcrumbs": breadcrumbs,
            "dataset": self.dataset,
            "dataurl": req.path_url.replace(".html", ""),
            "version": __version__,
        }

        return Response(body=template.render(context), headers=self.headers)

    def redirect(self, req):
        """Return a redirect to the ASCII response."""
        projection, selection = [], []
        for k in req.params:
            # selection
            if k.startswith("var1_") and req.params[k] != "--":
                name = k[5:]
                tokens = (
                    req.params[k],
                    req.params["op_%s" % name],
                    req.params["var2_%s" % name],
                )
                selection.append("".join(tokens))

            # projection
            if req.params[k] == "on":
                tokens = [k]
                i = 0
                while "%s[%d]" % (k, i) in req.params:
                    tokens.append("[%s]" % req.params["%s[%d]" % (k, i)])
                    i += 1
                projection.append("".join(tokens))

        # send to ASCII response
        location = "{0}.ascii?{1}&{2}".format(
            req.path_url[:-5], ",".join(projection), "&".join(selection)
        ).rstrip("?&")

        return HTTPSeeOther(location=location)



"""
From xarray.core.pycompat.py, under Apache 2.0 license.
"""

try:
    from contextlib import suppress
except ImportError:
    # Backport from CPython 3.5:
    # Used under the terms of Python's license, see licenses/PYTHON_LICENSE.

    class suppress:
        """Context manager to suppress specified exceptions
        After the exception is suppressed, execution proceeds with the next
        statement following the with statement.
             with suppress(FileNotFoundError):
                 os.remove(somefile)
         # Execution still resumes here if the file was already removed
        """

        def __init__(self, *exceptions):
            self._exceptions = exceptions

        def __enter__(self):
            pass

        def __exit__(self, exctype, excinst, exctb):
            # Unlike isinstance and issubclass, CPython exception
            # handling
            # currently only looks at the concrete type hierarchy (ignoring
            # the instance and subclass checking hooks). While Guido considers
            # that a bug rather than a feature, it's a fairly hard one
            # to fix
            # due to various internal implementation details. suppress
            # provides
            # the simpler issubclass based semantics, rather than trying
            # to
            # exactly reproduce the limitations of the CPython
            # interpreter.
            #
            # See http://bugs.python.org/issue12029 for more details
            return exctype is not None and issubclass(exctype, self._exceptions)



"""Test pydap namespace."""

import unittest

import pydap


class TestNamespace(unittest.TestCase):
    """Test pydap namespace."""

    def test_namespace(self):
        """Test the namespace."""
        self.assertEqual(pydap.__name__, "pydap")



"""Test the server-side functions that come with pydap."""

import copy
import unittest

from webtest import TestApp as App

from pydap.exceptions import ConstraintExpressionError, ServerError
from pydap.handlers.lib import BaseHandler
from pydap.tests.datasets import SimpleGrid, SimpleSequence
from pydap.wsgi.ssf import ServerSideFunctions


class TestDensity(unittest.TestCase):
    """Test the density function."""

    def setUp(self):
        """Create simple WSGI app."""
        self.app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))

    def test_wrong_type(self):
        """Test passing the wrong type."""
        app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))
        with self.assertRaises(ConstraintExpressionError):
            app.get("/.dds?density(SimpleGrid,SimpleGrid,SimpleGrid)")

    def test_plain(self):
        """Test a direct request."""
        res = self.app.get("/.asc")
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            '"1", 100, -10, 0, -1, 21, 35, 0\n'
            '"2", 200, 10, 500, 1, 15, 35, 100\n'
            "\n",
        )

    def test_projection(self):
        """Test using density as a projection."""
        res = self.app.get(
            "/.asc?density(cast.salinity,cast.temperature,cast.pressure)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Float64 rho;\n"
            "    } result;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "result.rho\n"
            "1024.37\n"
            "1026.29\n"
            "\n",
        )

    def test_selection(self):
        """Test using density as a selection."""
        res = self.app.get(
            "/.asc?cast.temperature&"
            "density(cast.salinity,cast.temperature,cast.pressure)>1025"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Int32 temperature;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.temperature\n"
            "15\n"
            "\n",
        )


class TestBounds(unittest.TestCase):
    """Test the ``bounds`` function, used by GrADS."""

    def setUp(self):
        """Create a simple WSGI app."""
        self.app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))

    def test_wrong_type(self):
        """Test passing a wrong type to the function."""
        app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))
        with self.assertRaises(ConstraintExpressionError):
            app.get(
                "/.dds?SimpleGrid"
                "&bounds(0,360,-90,90,500,500,00Z01JAN1970,00Z01JAN1970)"
            )

    def test_default(self):
        """Test the default bounding box."""
        res = self.app.get(
            "/.asc?cast&bounds(0,360,-90,90,0,500," "00Z01JAN1970,00Z01JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            "\n",
        )

    def test_selection_only(self):
        """Test using the function alone in the selection."""
        res = self.app.get(
            "/.asc?bounds(0,360,-90,90,0,500," "00Z01JAN1970,00Z01JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            "\n",
        )

    def test_subset(self):
        """Test a requesting a subset of the data."""
        res = self.app.get(
            "/.asc?bounds(0,360,-90,90,0,500," "00Z01JAN1970,00Z31JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            '"2", 200, 10, 500, 1, 15, 35, 100\n'
            "\n",
        )

    def test_subset_with_selection(self):
        """Test combining selections."""
        res = self.app.get(
            "/.asc?"
            "bounds(0,360,-90,90,0,500,"
            "00Z01JAN1969,00Z31JAN1970)&cast.lat<0"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            '"1", 100, -10, 0, -1, 21, 35, 0\n'
            "\n",
        )

    def test_projection(self):
        """Test bounds used as a projection."""
        res = self.app.get(
            "/.asc?cast.pressure&"
            "bounds(0,360,-90,90,0,500,"
            "00Z01JAN1970,00Z31JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.pressure\n"
            "100\n"
            "\n",
        )

    def test_point(self):
        """Test a request for a point."""
        res = self.app.get(
            "/.asc?bounds(100,100,-10,-10,0,0," "00Z31DEC1969,00Z31DEC1969)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        String id;\n"
            "        Int32 lon;\n"
            "        Int32 lat;\n"
            "        Int32 depth;\n"
            "        Int32 time;\n"
            "        Int32 temperature;\n"
            "        Int32 salinity;\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.id, cast.lon, cast.lat, cast.depth, cast.time, "
            "cast.temperature, cast.salinity, cast.pressure\n"
            '"1", 100, -10, 0, -1, 21, 35, 0\n'
            "\n",
        )

    def test_grads_step(self):
        """Test different GrADS time steps."""
        modified = copy.copy(SimpleSequence)
        modified.cast.time.attributes["grads_step"] = "1mn"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        res = app.get(
            "/.asc?cast.pressure&"
            "bounds(0,360,-90,90,0,500,12Z01JAN1970,12Z01JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.pressure\n"
            "\n",
        )

        modified.cast.time.attributes["grads_step"] = "1hr"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        res = app.get(
            "/.asc?cast.pressure&"
            "bounds(0,360,-90,90,0,500,12Z01JAN1970,12Z01JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.pressure\n"
            "\n",
        )

        modified.cast.time.attributes["grads_step"] = "1dy"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        res = app.get(
            "/.asc?cast.pressure&"
            "bounds(0,360,-90,90,0,500,12Z01JAN1970,12Z01JAN1970)"
        )
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Sequence {\n"
            "        Int32 pressure;\n"
            "    } cast;\n"
            "} SimpleSequence;\n"
            "---------------------------------------------\n"
            "cast.pressure\n"
            "100\n"
            "\n",
        )

        modified.cast.time.attributes["grads_step"] = "1mo"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        with self.assertRaises(NotImplementedError):
            res = app.get(
                "/.asc?cast.pressure&"
                "bounds(0,360,-90,90,0,500,"
                "12Z01JAN1970,12Z01JAN1970)"
            )

        modified.cast.time.attributes["grads_step"] = "1yr"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        with self.assertRaises(NotImplementedError):
            res = app.get(
                "/.asc?cast.pressure&"
                "bounds(0,360,-90,90,0,500,"
                "12Z01JAN1970,12Z01JAN1970)"
            )

        modified.cast.time.attributes["grads_step"] = "1xx"
        app = App(ServerSideFunctions(BaseHandler(modified)))
        with self.assertRaises(ServerError):
            res = app.get(
                "/.asc?cast.pressure&"
                "bounds(0,360,-90,90,0,500,"
                "12Z01JAN1970,12Z01JAN1970)"
            )


class TestMean(unittest.TestCase):
    """Test the ``mean`` function."""

    def setUp(self):
        """Create a simple WSGI app."""
        self.app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))

    def test_wrong_type(self):
        """Test passing a wrong type to mean function."""
        app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))
        with self.assertRaises(ConstraintExpressionError):
            app.get("/.dds?mean(sequence)")

    def test_base_type(self):
        """Test mean of base types."""
        res = self.app.get("/.asc?mean(x)")
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Float64 x;\n"
            "} SimpleGrid;\n"
            "---------------------------------------------\n"
            "x\n"
            "1\n",
        )

    def test_grid_type(self):
        """Test mean of grid objects."""
        res = self.app.get("/.asc?mean(SimpleGrid)")
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Grid {\n"
            "        Array:\n"
            "            Float64 SimpleGrid[y = 3];\n"
            "        Maps:\n"
            "            Int32 y[y = 2];\n"
            "    } SimpleGrid;\n"
            "} SimpleGrid;\n"
            "---------------------------------------------\n"
            "SimpleGrid.SimpleGrid\n"
            "[0] 1.5\n"
            "[1] 2.5\n"
            "[2] 3.5\n"
            "\n"
            "SimpleGrid.y\n"
            "[0] 0\n"
            "[1] 1\n"
            "\n"
            "\n",
        )

    def test_nested(self):
        """Test nested function calls."""
        res = self.app.get("/.asc?mean(mean(SimpleGrid))")
        self.assertEqual(
            res.text,
            "Dataset {\n"
            "    Grid {\n"
            "        Array:\n"
            "            Float64 SimpleGrid;\n"
            "        Maps:\n"
            "    } SimpleGrid;\n"
            "} SimpleGrid;\n"
            "---------------------------------------------\n"
            "SimpleGrid.SimpleGrid\n"
            "2.5\n"
            "\n",
        )



"""Parsing related functions.

This module defines functions to parse DAP objects, including a base parser for
DDS and DAS responses.

"""

import ast
import operator
import re

from ..exceptions import ConstraintExpressionError
from ..lib import get_var, unquote


def parse_projection(input, protocol="dap2"):
    """Split a projection into items.

    The function takes into account server-side functions, and parse slices
    into Python slice objects.

    Returns a list of names and slices.

    """
    if protocol == "dap2":
        key = ","
    elif protocol == "dap4":
        key = ";"

    def tokenize(input):
        start = pos = count = 0
        for char in input:
            if char == "(":
                count += 1
            elif char == ")":
                count -= 1
            elif char == key and count == 0:
                yield input[start:pos]
                start = pos + 1
            pos += 1
        yield input[start:]

    def parse(token):
        if "(" not in token:
            token = token.split(".")
            token = [re.match(r"(.*?)(\[.*\])?$", part).groups() for part in token]
            token = [(name, parse_hyperslab(slice_ or "")) for (name, slice_) in token]
        return token

    return list(map(parse, tokenize(input)))


def parse_selection(expression, dataset):
    """Parse a selection expression into its elements.

    This function will parse a selection expression into three tokens: two
    variables or values and a comparison operator. Variables are returned as
    pydap objects from a given dataset, while values are parsed using
    ``ast.literal_eval``.

    """
    id1, op, id2 = re.split("(<=|>=|!=|=~|>|<|=)", expression, 1)

    op = {
        "<=": operator.le,
        ">=": operator.ge,
        "!=": operator.ne,
        "=": operator.eq,
        ">": operator.gt,
        "<": operator.lt,
    }[op]

    try:
        id1 = get_var(dataset, id1)
    except Exception:
        id1 = ast.literal_eval(id1)

    try:
        id2 = get_var(dataset, id2)
    except Exception:
        id2 = ast.literal_eval(id2)

    return id1, op, id2


def parse_ce(query_string, protocol="dap2"):
    """Extract the projection and selection from the QUERY_STRING.

        >>> parse_ce('a,b[0:2:9],c&a>1&b<2')  # doctest: +NORMALIZE_WHITESPACE
        ([[('a', ())], [('b', (slice(0, 10, 2),))], [('c', ())]],
                ['a>1', 'b<2'])
        >>> parse_ce('a>1&b<2')
        ([], ['a>1', 'b<2'])

    This function can also handle function calls in the URL, according to the
    DAP specification:

        >>> ce = 'time&bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)'
        >>> print(parse_ce(ce))  # doctest: +NORMALIZE_WHITESPACE
        ([[('time', ())]],
                ['bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)'])

        >>> ce = 'time,bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)'
        >>> print(parse_ce(ce))  # doctest: +NORMALIZE_WHITESPACE
        ([[('time', ())],
            'bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)'], [])
        >>> parse_ce('mean(g,0)')
        (['mean(g,0)'], [])
        >>> parse_ce('mean(mean(g.a,1),0)')
        (['mean(mean(g.a,1),0)'], [])

    Returns a tuple with the projection and the selection.

    """
    if protocol == "dap2":
        key = "&"
        if len(query_string) > 0 and query_string[:8] == "dap4.ce=":
            raise ConstraintExpressionError(
                "The Constraint Expression %s does not follow the DAP2 "
                "model specification" % query_string,
            )
    elif protocol == "dap4":
        key = "|"
        # remove `dap4.ce=` from query string if there is a query
        if len(query_string) > 0 and query_string[:8] != "dap4.ce=":
            raise ConstraintExpressionError(
                "The Constraint Expression %s does not follow the DAP4"
                "model specification " % query_string,
            )
        query_string = query_string[8:]
    tokens = [token for token in unquote(query_string).split(key) if token]
    if not tokens:
        projection = []
        selection = []
    elif re.search("<=|>=|!=|=~|>|<|=", tokens[0]):
        projection = []
        selection = tokens
    else:
        projection = parse_projection(tokens[0], protocol)
        selection = tokens[1:]

    return projection, selection


def parse_hyperslab(hyperslab):
    """Parse a hyperslab, returning a Python tuple of slices."""
    exprs = [expr for expr in hyperslab[1:-1].split("][") if expr]

    out = []
    for expr in exprs:
        tokens = list(map(int, expr.split(":")))
        start = tokens[0]
        step = 1

        if len(tokens) == 1:
            stop = start + 1
        elif len(tokens) == 2:
            stop = tokens[1] + 1
        elif len(tokens) == 3:
            step = tokens[1]
            stop = tokens[2] + 1
        else:
            raise ConstraintExpressionError("Invalid hyperslab %s" % hyperslab)

        out.append(slice(start, stop, step))

    return tuple(out)


class SimpleParser(object):
    """A very simple parser."""

    def __init__(self, input, flags=0):
        self.buffer = input
        self.flags = flags

    def peek(self, regexp):
        """Check if a token is present and return it."""
        p = re.compile(regexp, self.flags)
        m = p.match(self.buffer)
        if m:
            token = m.group()
        else:
            token = ""
        return token

    def consume(self, regexp):
        """Consume a token from the buffer and return it."""
        p = re.compile(regexp, self.flags)
        m = p.match(self.buffer)
        if m:
            token = m.group()
            self.buffer = self.buffer[len(token) :]
        else:
            raise Exception("Unable to parse token: %s" % self.buffer[:10])
        return token



"""Test the DAP handler, which forms the core of the client."""

import csv

import numpy as np
import pytest

from pydap.handlers.csv import CSVHandler
from pydap.handlers.dap import DAPHandler


@pytest.fixture(scope="session")
def simple_data():
    data = [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]
    return data


@pytest.fixture(scope="session")
def simple_data_file(tmpdir_factory, simple_data):
    temp_file = str(tmpdir_factory.mktemp("data").join("simple_data.csv"))
    with open(temp_file, "w") as f:
        writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
        writer.writerow(["index", "temperature", "site"])
        for row in simple_data:
            writer.writerow(row)
    return temp_file


def test_open(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = DAPHandler("http://localhost:8001/", CSVHandler(simple_data_file)).dataset
    seq = dataset["sequence"]
    dtype = [("index", "<i4"), ("temperature", "<f8"), ("site", "S40")]
    retrieved_data = [line for line in seq]

    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=dtype), np.array(simple_data, dtype=dtype)
    )


def test_combined_slice(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = CSVHandler(simple_data_file).dataset
    seq = dataset["sequence"]
    retrieved_data = [line for line in seq[["temperature", "site"]][seq["index"] > 10]]

    dtype = [("temperature", "<f8"), ("site", "S40")]

    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=dtype),
        np.array([item[1:] for item in simple_data[1:]], dtype=dtype),
    )


def test_constrained(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = CSVHandler(simple_data_file).dataset
    seq = dataset["sequence"]
    retrieved_data = [line for line in seq[seq["index"] > 10]["site"][::2]]

    dtype = "S40"

    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=dtype),
        np.array([simple_data[idx][-1] for idx in [1, 3]], dtype=dtype),
    )



"""Fundamental functions for pydap responses.

pydap responses are WSGI applications that convert a dataset into different
representations, like the DDS, DAS and DODS responses described in the DAP
specification.

In addition to the official responses, pydap also has responses that generate
KML, WMS, JSON, etc., installed as third-party Python packages that declare the
"pydap.response" entry point.

"""

from importlib.metadata import entry_points

from ..lib import __version__
from ..model import DatasetType


def load_responses():
    """Load all available responses from the system, returning a dictionary."""
    eps = entry_points(group="pydap.response")
    Rs = [r for r in eps if r.module[:5] == "pydap"]
    nRs = [r for r in eps if r.module[:5] != "pydap"]
    base_dict = dict((r.name, r.load()) for r in Rs)

    opts_dict = dict((r.name, r.load()) for r in nRs)
    base_dict.update(opts_dict)
    return base_dict


class BaseResponse(object):
    """A base class for pydap responses.

    A pydap response is a WSGI application that converts a dataset into any
    other representation. The most know responses are the DDS, DAS and DODS
    responses from the DAP spec, which describe the dataset structure,
    attributes and data, respectively.

    According to the WSGI specification, WSGI applications must returned an
    iterable object when called. While this is traditionally a list of strings
    representing an HTML response, this is not the case for pydap. pydap will
    return an object (the response instance itself), which is an iterable that
    yields the corresponding output (a DDS response, eg).

    In practice, this means that the generation of the response is delayed
    until the data is being sent to the client. But since the response object
    also carries the original dataset, this means it's possible to write WSGI
    middleware that modifies the dataset directly. A WSGI middleware can add
    additional metadata to a dataset, eg, by adding attributes directly to the
    dataset object, without having to generate a new response.

    """

    def __init__(self, dataset):
        self.dataset = dataset
        self.headers = [
            ("OPeNDAP-Server", "pydap/%s" % __version__),
        ]

    def __call__(self, environ, start_response):
        start_response("200 OK", self.headers)
        return self

    def x_wsgiorg_parsed_response(self, type):
        r"""Avoid serialization of datasets.

        This function will return the contained dataset if ``type`` is a
        ``pydap.model.DatasetType`` object. Based on this proposal:

            http://wsgi.readthedocs.org/en/latest/specifications/ \
                    avoiding_serialization.html

        """
        if type is DatasetType:
            return self.dataset

    def __iter__(self):
        raise NotImplementedError("Subclasses must implement __iter__")



"""
pydap client.

This module contains functions to access DAP servers. The most common use is to
open a dataset by its canonical URL, ie, without any DAP related extensions
like dds/das/dods/html. Here is an example:

    >>> from pydap.client import open_url
    >>> dataset = open_url("http://test.pydap.org/coads.nc")

This will return a `DatasetType` object, which is a container for lazy
evaluated objects. Data is downloaded automatically when arrays are sliced or
when sequences are iterated.

It is also possible to download data directly from a dods (binary) response.
This allows calling server-specific functions, like those supported by the
Ferret and the GrADS data servers:

    >>> from pydap.client import open_dods_url
    >>> dataset = open_dods_url(
    ...     "http://test.pydap.org/coads.nc.dods",
    ...     metadata=True)

Setting the `metadata` flag will also request the das response, populating the
dataset with the corresponding metadata.

If the dods response has already been downloaded, it is possible to open it as
if it were a remote dataset. Optionally, it is also possible to specify a das
response:

    >>> from pydap.client import open_file
    >>> dataset = open_file(
    ...     "/path/to/file.dods", "/path/to/file.das")  #doctest: +SKIP

Remote datasets opened with `open_url` can call server functions. pydap has a
lazy mechanism for function call, supporting any function. Eg, to call the
`geogrid` function on the server:

    >>> dataset = open_url(
    ...     'http://test.opendap.org/dap/data/nc/coads_climatology.nc')
    >>> new_dataset = dataset.functions.geogrid(dataset.SST, 10, 20, -10, 60)
    >>> print(new_dataset.SST.SST.shape) #doctest: +SKIP
    (12, 12, 21)

"""

from io import BytesIO, open

import numpy
from requests.utils import urlparse, urlunparse

import pydap.handlers.dap
import pydap.lib
import pydap.model
import pydap.net
import pydap.parsers.das
import pydap.parsers.dds
import pydap.parsers.dmr
from pydap.lib import DEFAULT_TIMEOUT as DEFAULT_TIMEOUT


def open_url(
    url,
    application=None,
    session=None,
    output_grid=True,
    timeout=DEFAULT_TIMEOUT,
    verify=True,
    user_charset="ascii",
    protocol=None,
):
    """
    Open a remote URL, returning a dataset.

    set output_grid to `False` to retrieve only main arrays and
    never retrieve coordinate axes.
    """
    handler = pydap.handlers.dap.DAPHandler(
        url,
        application,
        session,
        output_grid,
        timeout=timeout,
        verify=verify,
        user_charset=user_charset,
        protocol=protocol,
    )
    dataset = handler.dataset

    # attach server-side functions
    dataset.functions = Functions(url, application, session, timeout=timeout)

    return dataset


def open_file(file_path, das_path=None):
    extension = file_path.split(".")[-1]
    if extension == "dods":
        return open_dods_file(file_path=file_path, das_path=das_path)
    elif extension == "dap":
        return open_dap_file(file_path=file_path, das_path=das_path)
    elif extension == "dds":
        pass
    elif extension == "dmr":
        return open_dmr_file(file_path=file_path)


def get_dmr_length(file_path):
    with open(file_path, "rb") as f:
        # First two bytes are CRLF
        if f.peek()[0:2] == b"\x04\x00":
            f.seek(2)
            dmr_len = numpy.frombuffer(f.read(2), dtype=">u2")[0]
        else:
            dap = f.read()
            dmr = b"<?xml" + dap.split(b"<?xml")[1]
            dmr = dmr.split(b"</Dataset>")[0] + b"</Dataset>\n\r\n"
            dmr_len = len(dmr)
    return dmr_len


def open_dmr_file(file_path):
    dmr_len = get_dmr_length(file_path)
    with open(file_path, "rb") as f:
        if f.peek()[0:2] == b"\x04\x00":
            # First 2 bytes are CRLF, second two bytes give the length of the
            # DMR; we skip over them
            f.seek(4)
            # We read the DMR minus the CRLF and newline (3 bytes)
        dmr = f.read(dmr_len)
    dmr = dmr.decode("ascii")
    dataset = pydap.parsers.dmr.dmr_to_dataset(dmr)
    return dataset


def open_dap_file(file_path, das_path=None):
    """Open a file downloaded from a `.dap` (dap4) response, retunring a
    dataset.

    Optionally, read also the `.das` response to assign attributes to the
    dataset.
    """

    dataset = open_dmr_file(file_path)

    with open(file_path, "rb") as f:
        dmr_len = get_dmr_length(file_path)
        # if f.peek()[0:2] == b'\x04\x00':
        f.seek(dmr_len)
        _ = f.read(4)
        pydap.handlers.dap.unpack_dap4_data(f, dataset)
    return dataset


def open_dods_file(file_path, das_path=None):
    """Open a file downloaded from a `.dods` (dap2) response, returning a
    dataset.

    Optionally, read also the `.das` response to assign attributes to the
    dataset.

    """
    dds = ""
    # This file contains both ascii _and_ binary data
    # Let's handle them separately in sequence
    # Without ignoring errors, the IO library will
    # actually read past the ascii part of the
    # file (despite our break from iteration) and
    # will error out on the binary data
    with open(
        file_path, "rt", buffering=1, encoding="ascii", newline="\n", errors="ignore"
    ) as f:
        for line in f:
            if line.strip() == "Data:":
                break
            dds += line
    dataset = pydap.parsers.dds.dds_to_dataset(dds)
    pos = len(dds) + len("Data:\n")

    with open(file_path, "rb") as f:
        f.seek(pos)
        dataset.data = pydap.handlers.dap.unpack_dap2_data(f, dataset)

    if das_path is not None:
        with open(das_path) as f:
            das = pydap.parsers.das.parse_das(f.read())
            pydap.parsers.das.add_attributes(dataset, das)

    return dataset


def open_dods_url(
    url,
    metadata=False,
    application=None,
    session=None,
    timeout=DEFAULT_TIMEOUT,
    verify=True,
):
    """Open a `.dods` response directly, returning a dataset."""

    r = pydap.net.GET(url, application, session, timeout=timeout)
    pydap.net.raise_for_status(r)

    dds, data = r.body.split(b"\nData:\n", 1)
    dds = dds.decode(r.content_encoding or "ascii")
    dataset = pydap.parsers.dds.dds_to_dataset(dds)
    stream = pydap.handlers.dap.StreamReader(BytesIO(data))
    dataset.data = pydap.handlers.dap.unpack_dap2_data(stream, dataset)

    if metadata:
        scheme, netloc, path, params, query, fragment = urlparse(url)
        dasurl = urlunparse(
            (scheme, netloc, path[:-4] + "das", params, query, fragment)
        )
        r = pydap.net.GET(dasurl, application, session, timeout=timeout, verify=verify)
        pydap.net.raise_for_status(r)
        das = pydap.parsers.das.parse_das(r.text)
        pydap.parsers.das.add_attributes(dataset, das)

    return dataset


class Functions(object):
    """Proxy for server-side functions."""

    def __init__(
        self, baseurl, application=None, session=None, timeout=DEFAULT_TIMEOUT
    ):
        self.baseurl = baseurl
        self.application = application
        self.session = session
        self.timeout = timeout

    def __getattr__(self, attr):
        return ServerFunction(
            self.baseurl, attr, self.application, self.session, timeout=self.timeout
        )


class ServerFunction(object):
    """A proxy for a server-side function.

    Instead of returning datasets, the function will return a proxy object,
    allowing nested requests to be performed on the server.

    """

    def __init__(
        self,
        baseurl,
        name,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
    ):
        self.baseurl = baseurl
        self.name = name
        self.application = application
        self.session = session
        self.timeout = timeout

    def __call__(self, *args):
        params = []
        for arg in args:
            if isinstance(arg, (pydap.model.DapType, ServerFunctionResult)):
                params.append(arg.id)
            else:
                params.append(pydap.lib.encode(arg))
        id_ = self.name + "(" + ",".join(params) + ")"
        return ServerFunctionResult(
            self.baseurl, id_, self.application, self.session, timeout=self.timeout
        )


class ServerFunctionResult(object):
    """A proxy for the result from a server-side function call."""

    def __init__(
        self,
        baseurl,
        id_,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
    ):
        self.id = id_
        self.dataset = None
        self.application = application
        self.session = session
        self.timeout = timeout

        scheme, netloc, path, params, query, fragment = urlparse(baseurl)
        self.url = urlunparse((scheme, netloc, path + ".dods", params, id_, None))

    def __getitem__(self, key):
        if self.dataset is None:
            self.dataset = open_dods_url(
                self.url, True, self.application, self.session, self.timeout
            )
        return self.dataset[key]

    def __getattr__(self, name):
        return self[name]


if __name__ == "__main__":
    fname = "/home/griessbaum/Dropbox/UCSB/pydap_cpt/pydap_notebooks/data/"
    ds = open_file(fname + "coads_climatology.nc.dmr")
    print(ds)



"""Server-side functions for pydap.

pydap maps function calls in the request URL to functions defined through entry
points. For example, pydap defines an entry point on the group "pydap.function"
with the name "density"; from ``setup.py``:

    [pydap.function]
    density = pydap.wsgi.functions.density

When a DAP client makes a request to the URL calling this function:

    ?cast.oxygen&density(cast.salt,cast.temp,cast.press)>1024

pydap will call the ``density`` function passing the dataset itself as the
first argument, followed by the variables from the sequence "cast".

New functions can be defined in third-party modules. Simply create a package
and declare the entry point in its ``setup.py`` file. The server will
automatically discover the new functions from the system.

"""

import re
from datetime import datetime, timedelta

import numpy as np

from ..exceptions import ConstraintExpressionError, ServerError
from ..lib import walk
from ..model import BaseType, GridType, SequenceType
from ..pycompat import suppress

with suppress(ImportError):
    import coards
    import gsw


def density(dataset, salinity, temperature, pressure):
    """Calculate in-situ density.

    This function calculated in-situ density from absolute salinity and
    conservative temperature, using the `gsw.rho` function. Returns a new
    sequence with the data.

    """
    # find sequence
    for sequence in walk(dataset, SequenceType):
        break
    else:
        raise ConstraintExpressionError(
            'Function "bounds" should be used on a Sequence.'
        )

    selection = sequence[salinity.name, temperature.name, pressure.name]
    rows = [tuple(row) for row in selection.iterdata()]
    data = np.rec.fromrecords(rows, names=["salinity", "temperature", "pressure"])
    rho = gsw.rho(data["salinity"], data["temperature"], data["pressure"])

    out = SequenceType("result")
    out["rho"] = BaseType("rho", units="kg/m**3")
    out.data = np.rec.fromrecords(rho.reshape(-1, 1), names=["rho"])
    return out


density.__version__ = "0.1"


def bounds(dataset, xmin, xmax, ymin, ymax, zmin, zmax, tmin, tmax):
    r"""Bound a sequence in space and time.

    This function is used by GrADS to access Sequences, eg:

        http://server.example.com/dataset.dods?sequence& \
                bounds(0,360,-90,90,500,500,00Z01JAN1970,00Z01JAN1970)

    We assume the dataset has only a single Sequence, which will be returned
    modified in place.

    """
    # find sequence
    for sequence in walk(dataset, SequenceType):
        break  # get first sequence
    else:
        raise ConstraintExpressionError(
            'Function "bounds" should be used on a Sequence.'
        )

    for child in sequence.children():
        if child.attributes.get("axis", "").lower() == "x":
            if xmin == xmax:
                sequence.data = sequence[child == xmin].data
            else:
                sequence.data = sequence[(child >= xmin) & (child <= xmax)].data
        elif child.attributes.get("axis", "").lower() == "y":
            if ymin == ymax:
                sequence.data = sequence[child == ymin].data
            else:
                sequence.data = sequence[(child >= ymin) & (child <= ymax)].data
        elif child.attributes.get("axis", "").lower() == "z":
            if zmin == zmax:
                sequence.data = sequence[child == zmin].data
            else:
                sequence.data = sequence[(child >= zmin) & (child <= zmax)].data
        elif child.attributes.get("axis", "").lower() == "t":
            start = datetime.strptime(tmin, "%HZ%d%b%Y")
            end = datetime.strptime(tmax, "%HZ%d%b%Y")
            units = child.attributes.get("units", "seconds since 1970-01-01")

            # if start and end are equal, add the step
            if start == end and "grads_step" in child.attributes:
                dt = parse_step(child.attributes["grads_step"])
                end = start + dt
                tmin = coards.format(start, units)
                tmax = coards.format(end, units)
                sequence.data = sequence[(child >= tmin) & (child < tmax)].data
            else:
                tmin = coards.format(start, units)
                tmax = coards.format(end, units)
                sequence.data = sequence[(child >= tmin) & (child <= tmax)].data

    return sequence


bounds.__version__ = "1.0"


def parse_step(step):
    """Parse a GrADS time step returning a timedelta."""
    value, units = re.search(r"(\d+)(.*)", step).groups()
    value = int(value)
    if units.lower() == "mn":
        return timedelta(minutes=value)
    if units.lower() == "hr":
        return timedelta(hours=value)
    if units.lower() == "dy":
        return timedelta(days=value)
    if units.lower() == "mo":
        raise NotImplementedError("Need to implement month time step")
    if units.lower() == "yr":
        raise NotImplementedError("Need to implement year time step")
    raise ServerError('Unknown units: "%s".' % units)


def mean(dataset, var, axis=0):
    """Calculate the mean of an array along a given axis.

    The input variable should be either a ``GridType`` or ``BaseType``. The
    function will return an object of the same type with the mean applied.

    """
    if not isinstance(var, (GridType, BaseType)):
        raise ConstraintExpressionError(
            'Function "mean" should be used on an array or grid.'
        )

    axis = int(axis)
    dims = tuple(dim for i, dim in enumerate(var.dimensions) if i != axis)

    # process basetype
    if isinstance(var, BaseType):
        return BaseType(
            name=var.name,
            data=np.mean(var.data[:], axis=axis),
            dimensions=dims,
            attributes=var.attributes,
        )

    # process grid
    out = GridType(name=var.name, attributes=var.attributes)
    out[var.array.name] = BaseType(
        name=var.array.name,
        data=np.mean(var.array.data[:], axis=axis),
        dimensions=dims,
        attributes=var.array.attributes,
    )
    for dim in dims:
        out[dim] = BaseType(
            name=dim,
            data=var[dim].data[:],
            dimensions=(dim,),
            attributes=var[dim].attributes,
        )
    return out


mean.__version__ = "1.0"



"""A DMR parser."""

import ast
import collections
import re
from xml.etree import ElementTree as ET

import numpy as np

import pydap.lib
import pydap.model

constructors = ("grid", "sequence", "structure")
name_regexp = r'[\w%!~"\'\*-]+'
dmr_atomic_types = (
    "Int8",
    "UInt8",
    "Byte",
    "Char",
    "Int16",
    "UInt16",
    "Int32",
    "UInt32",
    "Int64",
    "UInt64",
    "Float32",
    "Float64",
)

namespace = {"": "http://xml.opendap.org/ns/DAP/4.0#"}


def dap4_to_numpy_typemap(type_string):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = pydap.lib.DAP4_TO_NUMPY_PARSER_TYPEMAP[type_string]
    return np.dtype(dtype_str)


def get_variables(node, prefix="") -> dict:
    variables = collections.OrderedDict()
    group_name = node.get("name")
    if group_name is None:
        return variables
    if node.tag != "Dataset":
        prefix = prefix + "/" + group_name
    for subnode in node:
        if subnode.tag in dmr_atomic_types:
            name = subnode.get("name")
            if prefix != "":
                name = prefix + "/" + name
            variables[name] = {"element": subnode, "parent": node.tag}
        variables.update(get_variables(subnode, prefix))
    return variables


def get_named_dimensions(node, prefix=""):
    dimensions = {}
    group_name = node.get("name")
    if group_name is None:
        return dimensions
    if node.tag != "Dataset":
        prefix = prefix + "/" + group_name
    for subnode in node:
        if subnode.tag == "Dimension":
            name = subnode.get("name")
            if prefix != "":
                name = prefix + "/" + name
            dimensions[name] = int(subnode.attrib["size"])
        dimensions.update(get_named_dimensions(subnode, prefix))
    return dimensions


def get_dtype(element):
    dtype = element.tag
    dtype = dap4_to_numpy_typemap(dtype)
    return dtype


def get_atomic_attr(element):
    Float_types = ["Float32", "float", "Float64"]
    Int_types = ["Int16", "Int32", "Int64", "int", "Int8"]
    uInt_types = ["uInt16", "uInt32", "uInt64", "uint", "uInt8", "Char"]
    name = element.get("name")
    value = element.find("Value").text
    if value is None:
        # This could be because server is TDS.
        # If value is None still, then data is missing
        value = element.find("Value").get("value")
    _type = element.get("type")
    if _type in dmr_atomic_types and value is not None:
        if _type in Float_types:
            # keep float-type of value
            value = float(value)
        elif _type in Int_types or _type in uInt_types:
            # keeps integer-type of value
            value = int(value)
        else:
            try:
                value = ast.literal_eval(value)
            except ValueError:
                # leaves value as string
                raise Warning(
                    """
                    Pydap failed to retrieve Attribute: `{}` element during
                    parsing of the `DMR`.""".format(
                        name
                    )
                )
    return name, value


def get_attributes(element, attributes={}):
    attribute_elements = element.findall("Attribute")
    for attribute_element in attribute_elements:
        name, value = get_atomic_attr(attribute_element)
        attributes[name] = value
    return attributes


def get_dim_names(element):
    # Not to be confused with dimensions
    dimension_elements = element.findall("Dim")
    dimensions = []
    for dimension_element in dimension_elements:
        name = dimension_element.get("name")
        if name is None:
            # We might have unnamed dimensions
            return dimensions
        if name.find("/", 1) == -1:
            # If this is a root Dimension, we remove the leading slash
            name = name.replace("/", "")
        dimensions.append(name)
    return dimensions


def get_dim_sizes(element):
    dimension_elements = element.findall("Dim")
    dimension_sizes = ()
    for dimension_element in dimension_elements:
        name = dimension_element.get("name")
        if name is None:
            size = int(dimension_element.get("size"))
            dimension_sizes += (size,)
    return dimension_sizes


def get_maps(element):
    maps = element.findall("Map")
    Maps = tuple([maps[i].get("name") for i in range(len(maps))])
    return Maps


def get_groups(node, prefix="/"):
    groups = node.findall("Group")  # may be seveal elements
    out = {}
    for group in groups:
        fqname = prefix + group.attrib["name"]
        named_dimensions = get_named_dimensions(group)
        global_dimensions = []
        for name, size in named_dimensions.items():
            global_dimensions.append([name.split("/")[-1], size])
        out.update(
            {
                fqname: {
                    "attributes": get_attributes(group, {}),
                    "Maps": get_maps(group),
                    "dimensions": {k: v for k, v in global_dimensions},
                    "path": prefix,
                }
            }
        )
        subgroups = group.findall("Group")
        if len(subgroups) > 0:
            out.update(get_groups(group, fqname + "/"))
    return out


def dmr_to_dataset(dmr):
    """Return a dataset object from a DMR representation."""

    # Parse the DMR. First dropping the namespace
    dom_et = DMRParser(dmr).node
    # emtpy dataset
    if DMRParser(dmr).Groups:
        split_by = "/"
    else:
        split_by = None

    dataset = DMRParser(dmr).init_dataset()

    variables = get_variables(dom_et)
    named_dimensions = get_named_dimensions(dom_et)

    # get Global dimensions at root level
    global_dimensions = []
    for name, size in named_dimensions.items():
        if len(name.split(split_by)) == 1:
            global_dimensions.append([name, size])

    dataset.dimensions = {k: v for k, v in global_dimensions}

    # Add size entry for dimension variables
    for name, size in named_dimensions.items():
        if name in variables:
            variables[name]["size"] = size

    # Bootstrap variables
    for name, variable in variables.items():
        variable["name"] = name
        variable["attributes"] = get_attributes(variable["element"], {})
        variable["dtype"] = get_dtype(variable["element"])
        variable["dims"] = get_dim_names(variable["element"])
        variable["maps"] = get_maps(variable["element"])
        variable["shape"] = get_dim_sizes(variable["element"])

    # Add shape element to variables
    for name, variable in variables.items():
        for dim in variable["dims"]:
            if dim in variables.items():
                variable["shape"] += (variables[dim]["size"],)
            else:
                variable["shape"] += (named_dimensions[dim],)

    for name, variable in variables.items():
        var_name = variable["name"]
        path = None
        if len(var_name.split(split_by)) > 1:
            # path-like name - Groups!
            parts = var_name.split(split_by)
            var_name = parts[-1]
            path = ("/").join(parts[:-1])
            variable["attributes"]["path"] = path

        data = DummyData(dtype=variable["dtype"], shape=variable["shape"], path=path)
        # make sure all dimensions have qualifying name
        Dims = []
        for dim in variable["dims"]:
            if len(dim.split(split_by)) == 1:
                Dims.append("/" + dim)
            else:
                Dims.append(dim)
        # pass along maps
        var_kwargs = {
            "name": name,
            "data": data,
            "dimensions": Dims,
            "attributes": variable["attributes"],
        }
        if "maps" in variable.keys():
            var_kwargs.update({"Maps": variable["maps"]})
        if "parent" in variable.keys() and variable["parent"] in [
            "Sequence",
            "Structure",
        ]:
            parts = name.split(split_by)
            parent_name = parts[-2]
            path = ("/").join(parts[:-2])
            if (
                variable["parent"] == "Sequence"
                and parent_name not in dataset[path].keys()
            ):
                dataset.createSequence(("/").join(parts), path=path)
            elif (
                variable["parent"] == "Structure"
                and parent_name not in dataset[path].keys()
            ):
                dataset.createStructure(("/").join(parts), path=path)
        else:
            dataset.createVariable(**var_kwargs)

    return dataset


class DMRParser(object):
    """A parser for the DMR."""

    def __init__(self, dmr):
        self.dmr = dmr
        self.Groups = None

        _dmr = re.sub(' xmlns="[^"]+"', "", self.dmr, count=1)
        self.node = ET.fromstring(_dmr)
        if len(get_groups(self.node)) > 0:
            # no groups here
            self.Groups = True

    def init_dataset(self):
        """creates an empty dataset with a name and attributes"""
        dataset_name = self.node.get("name")
        dataset = pydap.model.DatasetType(dataset_name)
        AttsNames = [subnode.get("name") for subnode in self.node.findall("Attribute")]
        Attrs = {}
        for subnode in self.node:
            if subnode.get("name") in AttsNames:
                if subnode.get("type") not in dmr_atomic_types + (
                    "String",
                    "URI",
                ):
                    # container type
                    Attrs = get_attributes(subnode, Attrs)
                else:
                    name, value = get_atomic_attr(subnode)
                    Attrs[name] = value
        dataset.attributes = Attrs
        # create Groups via dict
        Groups = get_groups(self.node)
        for key in Groups:
            dims = Groups[key].pop("dimensions", None)
            Maps = Groups[key].pop("Maps", None)
            attributes = Groups[key].pop("attributes", None)
            dataset.createGroup(
                name=key, dimensions=dims, Maps=Maps, attributes=attributes
            )

        return dataset


class DummyData(object):
    def __init__(self, dtype, shape, path):
        self.dtype = dtype
        self.shape = shape
        self.path = path



"""pydap handler for NetCDF3/4 files."""

import os
import re
import time
from collections import OrderedDict
from email.utils import formatdate
from stat import ST_MTIME

import numpy as np

from pydap.exceptions import OpenFileError
from pydap.handlers.lib import BaseHandler
from pydap.model import BaseType, DatasetType
from pydap.pycompat import suppress

# Check for netCDF4 presence:
with suppress(ImportError):
    try:
        from netCDF4 import Dataset as netcdf_file

        def attrs(var):
            return dict((k, getattr(var, k)) for k in var.ncattrs())

    except ImportError:
        from scipy.io.netcdf import netcdf_file

        def attrs(var):
            return var._attributes


class NetCDFHandler(BaseHandler):
    """A simple handler for NetCDF files.
    Here's a standard dataset for testing sequential data:
    """

    extensions = re.compile(r"^.*\.(nc4|nc|cdf)$", re.IGNORECASE)

    def __init__(self, filepath):
        BaseHandler.__init__(self)

        self.filepath = filepath
        try:
            with netcdf_file(self.filepath, "r") as source:
                self.additional_headers.append(
                    (
                        "Last-modified",
                        (
                            formatdate(
                                time.mktime(time.localtime(os.stat(filepath)[ST_MTIME]))
                            )
                        ),
                    )
                )

                # shortcuts
                vars = source.variables
                dims = source.dimensions

                # Add dimensions when creating the DatasetType
                Dims = OrderedDict()
                fqn_dims = OrderedDict()  # keep track of fully qualifying names of dims
                for dim in dims:
                    fqn_dims.update({"/" + dim: dim})
                    if dims[dim] is None:
                        self.dataset.attributes["DODS_EXTRA"] = {
                            "Unlimited_Dimension": dim,
                        }
                    else:
                        Dims.update({dims[dim].name: dims[dim].size})
                # build dataset

                name = os.path.split(filepath)[1]
                self.dataset = DatasetType(
                    name, dimensions=Dims, attributes=dict(NC_GLOBAL=attrs(source))
                )

                # add grids
                grids = [var for var in vars if var not in dims]
                for grid in grids:
                    # make dimension a fully qualifying name
                    dimensions = tuple(["/" + dim for dim in vars[grid].dimensions])
                    self.dataset[grid] = BaseType(
                        grid,
                        LazyVariable(source[grid], grid, grid, self.filepath),
                        dimensions=dimensions,
                        **attrs(vars[grid]),
                    )

                if len(source.groups) > 0:
                    # start at root level
                    path = source.path
                    for vdim in source.dimensions:
                        fqn_dims.update({path + vdim: vdim})  # fqn is unique
                    fqn_dims = group_fqn(self.dataset, source, self.filepath, fqn_dims)

                vdims = [dim for dim in dims if dim in vars]
                for dim in vdims:
                    data = vars[dim][:]
                    attributes = attrs(vars[dim])
                    self.dataset[dim] = BaseType(dim, data, None, attributes)
        except Exception as exc:
            raise
            message = "Unable to open file %s: %s" % (filepath, exc)
            raise OpenFileError(message)


def group_fqn(_dataset, _source, _filepath, _fqn_dims=OrderedDict()):
    """Function to create nested DAP objects with fully-qualified-names
    within a hierarchy. Returns a dictionary with mapping between dimension
    names used at the group/array level and their fully qualifying names.

    Parameters:
    ----------
        dataset: <pydap DatasetType>
            dataset to `write` object to.
        _source: <class netCDF4 >
            root, group, subgroup.
        _fqn_dims: dict,
            dict to create mapping between dimension names
            and their full-qualyfing names.

    Returns:
        _fqn_dims: dict
            Updated dict with mapping `dim_name` <--> fqn
    """
    for group in _source.groups:
        _path = _source[group].path
        if _path[-1] != "/":
            _path = _path + "/"
        # create group and attrs + dims (non-fqn)
        dims = _source[group].dimensions
        Dims = OrderedDict()
        for dim in dims:
            if dim not in _fqn_dims.items():
                _fqn_dims.update({_path + dim: dim})
            Dims.update(
                {
                    _source[group]
                    .dimensions[dim]
                    .name: _source[group]
                    .dimensions[dim]
                    .size
                }
            )
        _attrs = dict(
            (attr, _source[group].getncattr(attr)) for attr in _source[group].ncattrs()
        )

        _dataset.createGroup(_source[group].path, dimensions=Dims, **_attrs)
        # now vars
        Vars = _source[group].variables
        for var in Vars:
            data = _source[group][var]
            dims = list(_source[group][var].dimensions)  # these must have fqn
            vdims = []  # create mapping for fqn
            for dim in dims:
                for key, value in _fqn_dims.items():
                    if dim == value:
                        match_dim = key  # overwrites until the last one (most recent)
                vdims.append(match_dim)
            vattrs = dict(
                (attr, _source[group][var].getncattr(attr))
                for attr in _source[group][var].ncattrs()
            )
            _dataset.createVariable(
                _path + var,
                data=LazyVariable(data, var, _path + var, _filepath),
                dimensions=tuple(vdims),
                **vattrs,
            )
        # check if there are nested group
        if len(_source[group].groups) > 0:
            _fqn_dims = group_fqn(_dataset, _source[group], _filepath, _fqn_dims)
    return _fqn_dims


class LazyVariable:
    def __init__(self, var, name, path, filepath):
        self.filepath = filepath
        self.path = path
        # var = source[self.path]
        # self.dimensions = fqn_dims
        self.dtype = np.dtype(var.dtype)
        self.datatype = var.dtype
        self.ndim = len(var.dimensions)
        self._shape = var.shape
        self._reshape = var.shape
        self.scale = True
        self.name = name
        self.size = np.prod(self.shape)
        self._attributes = dict((attr, var.getncattr(attr)) for attr in var.ncattrs())
        return

    def chunking(self):
        return "contiguous"

    def filters(self):
        return None

    def get_var_chunk_cache(self):
        raise NotImplementedError("get_var_chunk_cache is not implemented")
        return

    def ncattrs(self):
        return self._attributes

    def getncattr(self, attr):
        return self._attributes[attr]

    def __getattr__(self, name):
        # from netcdf4-python
        # if name in _private_atts, it is stored at the python
        # level and not in the netCDF file.
        if name.startswith("__") and name.endswith("__"):
            # if __dict__ requested, return a dict with netCDF attributes.
            if name == "__dict__":
                names = self.ncattrs()
                values = []
                for name in names:
                    values.append(self._attributes[name])
                return OrderedDict(zip(names, values))
            else:
                raise AttributeError
        else:
            return self.getncattr(name)

    def getValue(self):
        return self[...]

    def __array__(self):
        return self[...]

    def __getitem__(self, key):
        with netcdf_file(self.filepath, "r") as source:
            # Avoid applying scale_factor, see
            # https://github.com/pydap/pydap/issues/190
            source.set_auto_scale(False)
            return (
                np.asarray(source[self.path][key])
                .astype(self.dtype)
                .reshape(self._reshape)
            )

    def reshape(self, *args):
        if len(args) > 1:
            self._reshape = args
        else:
            self._reshape = args
        return self

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        if not self.shape:
            raise TypeError("len() of unsized object")
        else:
            return self.shape[0]

    def _getdims(self):
        return self.dimensions


if __name__ == "__main__":
    import sys

    from werkzeug.serving import run_simple

    application = NetCDFHandler(sys.argv[1])
    run_simple("localhost", 8001, application, use_reloader=True)



"""
Test the follow redirects and handling of more complex routing situations
"""

import requests
import requests_mock
from webob.request import Request

from pydap.net import create_request


def test_redirect():
    """Test that redirection is handled properly"""
    mock_url = "http://www.test.com/"
    # mock_url is redirected to https:
    redirect_url = "http://www.test2.com/"
    with requests_mock.Mocker() as m:
        m.register_uri("HEAD", mock_url, text="resp2", status_code=200)
        req = create_request(mock_url)
        assert len(m.request_history) == 1
        assert isinstance(req, Request)
        assert req.headers["Host"] == "www.test.com:80"

    # Without session:
    with requests_mock.Mocker() as m:
        m.register_uri(
            "HEAD",
            mock_url,
            text="resp1",
            status_code=301,
            headers={"Location": redirect_url},
        )
        m.register_uri("HEAD", redirect_url, text="resp2", status_code=200)
        req = create_request(mock_url)
        # Ensure follow redirect:
        assert len(m.request_history) == 2
        assert isinstance(req, Request)
        assert req.headers["Host"] == "www.test2.com:80"

    # With session:
    with requests_mock.Mocker() as m:
        m.register_uri(
            "HEAD",
            mock_url,
            text="resp1",
            status_code=301,
            headers={"Location": redirect_url},
        )
        m.register_uri("HEAD", redirect_url, text="resp2", status_code=200)
        req = create_request(mock_url, session=requests.Session())
        # Ensure follow redirect:
        assert len(m.request_history) == 2
        assert isinstance(req, Request)
        assert req.headers["Host"] == "www.test2.com:80"



"""Test the data model."""

import copy
import warnings
from collections.abc import Mapping

import numpy as np
import pytest

from pydap.model import (
    BaseType,
    DapType,
    DatasetType,
    GridType,
    GroupType,
    SequenceType,
    StructureType,
)

warnings.simplefilter("always")


# Test the super class for pydap types.
def test_DapType_quote():
    """Test that names are properly quoted."""
    var = DapType("foo.bar[0]")
    assert var.name == "foo%2Ebar%5B0%5D"


def test_DapType_attributes():
    """Test attribute assignment.

    Note that ``**kwargs`` have precedence.

    """
    attributes = {"foo": "bar", "value": 43}
    var = DapType("var", attributes=attributes, value=42)
    assert var.attributes == {"foo": "bar", "value": 42}


def test_DapType_id():
    """Test id assignment."""
    var = DapType("var")
    assert var._id == "var"
    assert var.id == "var"


def test_DapType_repr():
    """Test the ``__repr__`` method."""
    var = DapType("var", foo="bar")
    assert repr(var) == "DapType('var', {'foo': 'bar'})"


def test_DapType_id_property():
    """Test id set/get property."""
    var = DapType("var")
    var.id = "new"
    assert var.id == "new"
    assert var._id == "new"


def test_DapType_getattr():
    """Test lazy attribute retrieval."""
    var = DapType("var", foo="bar", value=42)

    assert var.attributes["foo"] == "bar"
    assert var.foo == "bar"

    with pytest.raises(AttributeError):
        var.baz

    with pytest.raises(KeyError):
        var.attributes["baz"]


def test_DapType_children():
    """Test children method."""
    var = DapType("var")
    assert var.children() == ()


# Test the base pydap type.
def test_BaseType_no_data():
    """Test empty data and dimensions attributes."""
    var = BaseType("var")
    assert var.data is None
    assert var.dimensions == ()


def test_BaseType_data_and_dimensions():
    """Test data and dimensions assignment."""
    var = BaseType("var", [42], ("x",))
    assert var.data == [42]
    assert var.dimensions == ("x",)


def test_BaseType_repr():
    """Test ``__repr__`` method."""
    var = BaseType("var", 42, foo="bar")
    assert repr(var) == "<BaseType with data array(42)>"


def test_BaseType_dtype():
    """Test ``dtype`` property."""
    var = BaseType("var", np.array(1, np.int32))
    assert var.dtype == np.int32


def test_BaseType_shape():
    """Test ``shape`` property."""
    var = BaseType("var", np.arange(16).reshape(2, 2, 2, 2))
    assert var.shape == (2, 2, 2, 2)


def test_BaseType_size():
    """Test ``size`` property."""
    var = BaseType("var", np.arange(16).reshape(2, 2, 2, 2))
    assert var.size == 16


def test_BaseType_ndim():
    """Test ``ndim`` property."""
    var = BaseType("var", np.arange(16).reshape(2, 2, 2, 2))
    assert var.ndim == 4


def test_BaseType_copy():
    """Test lightweight ``__copy__`` method."""
    original = BaseType("var", np.array(1))
    clone = copy.copy(original)

    # note that clones share the same data:
    assert original is not clone
    assert original.data is clone.data

    # test attributes
    assert original.id == clone.id
    assert original.name == clone.name
    assert original.dimensions == clone.dimensions
    assert original.attributes == clone.attributes


def test_BaseType_comparisons():
    """Test that comparisons are applied to data."""
    var = BaseType("var", np.array(1))
    assert var == 1
    assert var != 2
    assert var >= 0
    assert var <= 2
    assert var > 0
    assert var < 2


def test_BaseType_sequence_protocol():
    """Test that the object behaves like a sequence."""
    var = BaseType("var", np.arange(10))
    assert var[-5] == 5
    assert len(var) == 10
    assert list(var) == list(range(10))


def test_BaseType_iter_protocol():
    """Test that the object behaves like an iterable."""
    var = BaseType("var", np.arange(10))
    assert list(iter(var)) == list(range(10))


def test_BaseType_array():
    """Test array conversion."""
    var = BaseType("var", np.arange(16).reshape(2, 2, 2, 2))
    np.testing.assert_array_equal(np.array(var), np.arange(16).reshape(2, 2, 2, 2))


def test_BaseType_itemsize():
    """Test itemsize of array"""
    var = BaseType("var", np.arange(4, dtype=np.int32))
    assert var.itemsize == np.arange(1, dtype=np.int32).itemsize


def test_BaseType_nbytes():
    var = BaseType("var", np.arange(4, dtype=np.int32))
    assert var.nbytes == np.arange(4, dtype=np.int32).nbytes


# Test pydap structures.
def test_StructureType_init():
    """Test attributes used for dict-like behavior."""
    var = StructureType("var")
    assert var._visible_keys == []
    assert var._dict == {}


def test_StructureType_instance():
    """Test that it is a Mapping and DapType."""
    var = StructureType("var")
    assert isinstance(var, Mapping)
    assert isinstance(var, DapType)


def test_StructureType_repr():
    """Test ``__repr__`` method."""
    var = StructureType("var")
    assert repr(var) == "<StructureType with children >"

    var["one"] = BaseType("one")
    var["two"] = BaseType("two")
    assert repr(var) == "<StructureType with children 'one', 'two'>"


def test_StructureType_len():
    """Test ``__len__`` method."""
    var = StructureType("var")
    var["one"] = BaseType("one")
    var["two"] = BaseType("two")
    assert len(var) == 2


def test_StructureType_contains():
    """Test container behavior."""
    var = StructureType("var")
    var["one"] = BaseType("one")
    assert "one" in var


def test_StructureType_lazy_attribute():
    """Test lazy attribute, returning first child."""
    var = StructureType("var", value=42, one="1")
    var["one"] = BaseType("one")
    assert var.value == 42
    assert var.one is var["one"]


def test_StructureType_children():
    """Test children iteration, should return all children."""
    var = StructureType("var", value=42, one="1")
    var["one"] = BaseType("one")
    var["two"] = BaseType("two")
    assert list(var.children()) == [var["one"], var["two"]]


def test_StructureType_setitem():
    """Test item assignment.

    Assignment requires the key and the name of the variable to be
    identical. It also takes care of reordering children that are
    reinserted.

    """
    var = StructureType("var")
    var["foo.bar"] = BaseType("foo.bar")
    assert list(var.keys()) == ["foo%2Ebar"]

    with pytest.raises(KeyError):
        var["bar"] = BaseType("baz")

    # test reordering
    var["bar"] = BaseType("bar")
    var["foo.bar"] = BaseType("foo.bar")
    assert list(var.keys()) == ["bar", "foo%2Ebar"]


def test_StructureType_getitem():
    """Test item retrieval."""
    var = StructureType("var")
    child = BaseType("child")
    child.data = np.array([[[0, 1]]])
    var["child"] = child
    assert var["child"] is child
    assert var["child.child"] is child
    with pytest.raises(KeyError):
        var["unloved child"]
    with pytest.raises(KeyError):
        var[:]

    assert var["parent.child"] is child
    assert var["grandparent.parent.child"] is child


def test_StructureType_getitem_tuple():
    """Test multiple item retrieval."""
    var = StructureType("var")
    for name in ["child1", "child2", "child3"]:
        child = BaseType(name)
        var[name] = child
        assert var[name] is child
    assert list(var["child1", "child3"].keys()) == ["child1", "child3"]
    assert list(var["child1", "child3"]._all_keys()) == ["child1", "child2", "child3"]
    with pytest.raises(KeyError):
        var["unloved child"]


def test_StructureType_delitem():
    """Test item deletion."""
    var = StructureType("var")
    var["one"] = BaseType("one")
    var["two"] = BaseType("two")
    var["three"] = BaseType("three")

    assert list(var.keys()) == ["one", "two", "three"]

    del var["one"]
    assert list(var.keys()) == ["two", "three"]

    # Make sure that one can safely delete
    # a non visible child:
    subset = var[("two",)]
    assert list(subset.keys()) == ["two"]
    assert isinstance(subset, StructureType)
    subset.__delitem__("three")

    # Cannot delete an inexistent child:
    with pytest.raises(KeyError):
        del var["inexistent"]


def test_StructureType_get_data():
    """Test that structure collects data from children."""
    var = StructureType("var", value=42, one="1")
    var["one"] = BaseType("one", 1)
    var["two"] = BaseType("two", 2)
    assert var.data == [1, 2]


def test_StructureType_set_data():
    """Test that data is propagated to children."""
    var = StructureType("var", value=42, one="1")
    var["one"] = BaseType("one")
    var["two"] = BaseType("two")
    var.data = [10, 20]
    assert var["one"].data == 10
    assert var["two"].data == 20


def test_StructureType_copy():
    """Test lightweight clone of a structure."""
    original = StructureType("var", value=42, one="1")
    original["one"] = BaseType("one")
    original["two"] = BaseType("two")
    original.data = [10, 20]

    clone = copy.copy(original)

    # note that clones share the same data:
    assert original is not clone
    assert original["one"] is not clone["one"]
    assert original["one"].data is clone["one"].data
    assert original["two"] is not clone["two"]
    assert original["two"].data is clone["two"].data

    # test attributes
    assert original.id == clone.id
    assert original.name == clone.name


# Test a pydap structure.
def test_DatasetType_setitem():
    """Test item assignment."""
    dataset = DatasetType("dataset")
    dataset["one"] = BaseType("one")
    assert dataset["one"].id == "one"


def test_DatasetType_id():
    """Test that the dataset id is not propagated."""
    dataset = DatasetType("dataset")
    child = BaseType("child")
    child.id = "error"
    dataset["child"] = child
    assert child.id == "child"


@pytest.fixture
def sequence_example():
    """Create a standard sequence from the DAP spec."""
    example = SequenceType("example")
    example["index"] = BaseType("index")
    example["temperature"] = BaseType("temperature")
    example["site"] = BaseType("site")
    example.data = np.rec.fromrecords(
        [
            (10, 15.2, "Diamond_St"),
            (11, 13.1, "Blacktail_Loop"),
            (12, 13.3, "Platinum_St"),
            (13, 12.1, "Kodiak_Trail"),
        ],
        names=list(example.keys()),
    )
    return example


def test_SequenceType_data(sequence_example):
    """Test data assignment in sequences."""
    np.testing.assert_array_equal(
        sequence_example["index"].data, np.array([10, 11, 12, 13])
    )


def test_SequenceType_len(sequence_example):
    """Test that length is read from the data attribute."""
    assert len(list(sequence_example.keys())) == 3
    assert len(sequence_example) == 4


def test_SequenceType_iterdata(sequence_example):
    """Test that data iteration happens over data."""
    for a, b in zip(sequence_example.iterdata(), sequence_example.data):
        for sub_a, sub_b in zip(a, b):
            assert sub_a == sub_b


def test_SequenceType_iter(sequence_example):
    """Test that iteration happens ove data."""
    # Remove in pydap 3.4
    for a, b in zip(iter(sequence_example), sequence_example.data):
        for sub_a, sub_b in zip(a, b):
            assert sub_a == sub_b


def test_SequenceType_items(sequence_example):
    """Test that iteration happens over the child names."""
    assert list(sequence_example.items()) == [
        (key, sequence_example[key]) for key in ["index", "temperature", "site"]
    ]


def test_SequenceType_values(sequence_example):
    """Test that iteration happens over the child names."""
    assert list(sequence_example.values()) == [
        sequence_example[key] for key in ["index", "temperature", "site"]
    ]


def test_SequenceType_getitem(sequence_example):
    """Test item retrieval.

    The ``__getitem__`` method is overloaded for sequences, and behavior
    will depend on the type of the key. It can either return a child or a
    new sequence.

    """
    # a string should return the corresponding child
    assert isinstance(sequence_example["index"], BaseType)

    # a tuple should reorder the children
    assert list(sequence_example.keys()) == ["index", "temperature", "site"]
    modified = sequence_example["site", "temperature"]
    assert isinstance(modified, SequenceType)
    assert list(modified.keys()) == ["site", "temperature"]

    # the return sequence is a new one
    assert sequence_example is not modified
    assert sequence_example["site"] is not modified["site"]
    assert isinstance(sequence_example["site"], BaseType)

    # and the data is not shared
    assert sequence_example["site"].data is not modified["site"].data

    subset = sequence_example["site"][::2]
    assert sequence_example is not subset
    assert isinstance(subset, BaseType)

    # it is also possible to slice the data, returning a new sequence
    subset = sequence_example[sequence_example["index"] > 11]
    assert sequence_example is not subset
    assert isinstance(subset, SequenceType)
    np.testing.assert_array_equal(
        subset.data, sequence_example.data[sequence_example.data["index"] > 11]
    )


def test_SequenceType_copy(sequence_example):
    """Test the lightweight ``__copy__`` method."""
    clone = copy.copy(sequence_example)
    assert sequence_example is not clone
    assert (sequence_example.data == clone.data).all()


def test_DatasetType_get_item_directory_path():
    dataset = DatasetType("dataset")
    dataset.createVariable("/child")
    assert dataset[""] == dataset
    assert dataset["/"] == dataset


def test_DatasetType_set_item_directory_path(sequence_example):
    dataset = DatasetType("dataset")
    # first create Groups, and then Sequence
    dataset.createGroup("/Group1")
    dataset.createGroup("/Group1/Group2")
    dataset.createSequence("/Group1/Group2/example")
    path = "/Group1/Group2/example"
    dataset[path] = sequence_example
    assert "Group1" in list(dataset.keys())
    assert "Group2" in list(dataset["/Group1"].keys())
    assert isinstance(dataset["/Group1"], GroupType)
    assert isinstance(dataset["/Group1/Group2"], GroupType)
    assert isinstance(dataset["/Group1/Group2/example"], SequenceType)


def test_DatasetType_fails_createDAPType():
    dataset = DatasetType("dataset")
    name = "/Group1/Group2"
    warning = "Failed to create `Group2` because parent `Group1` does not exist!"
    with pytest.warns(UserWarning, match=warning):
        dataset.createGroup(name)
    assert "Group1" not in dataset
    assert "Group2" not in dataset


def test_DatasetType_groups(sequence_example):
    dataset = DatasetType("dataset")
    dataset.createGroup("/Group1")
    dataset.createGroup("/Group1/SubGroup1")
    dataset.createGroup("/Group2")
    dataset.createGroup("/Group2/SubGroup2")
    groups = dataset.groups()
    values = {
        "Group1": "/",
        "SubGroup1": "/Group1/",
        "Group2": "/",
        "SubGroup2": "/Group2/",
    }
    assert groups == values


def test_DatasetType_sequences(sequence_example):
    dataset = DatasetType("dataset")
    dataset.createGroup("/Group1")
    dataset.createSequence("/Group1/example")
    dataset["/Group1/example"] = sequence_example
    seq_root_level = dataset.sequences()
    sqns = dataset["/Group1"].sequences()
    assert seq_root_level == sqns
    assert sqns == {"example": ["index", "temperature", "site"]}


def test_DatasetType_variables():
    dataset = DatasetType("dataset")
    dataset.createGroup("/Group1")
    dataset.createVariable("/root_variable", dtype=np.float32)
    dataset.createVariable("/Group1/other_variable", dtype=np.int64)
    assert dataset.variables() == {"root_variable": np.dtype("float32")}
    assert dataset["/Group1"].variables() == {"other_variable": np.dtype("int64")}


# Test pydap grids.
@pytest.fixture()
def gridtype_example():
    """Create a simple grid."""
    example = GridType("example")
    example["a"] = BaseType("a", data=np.arange(30 * 50).reshape(30, 50))
    example["x"] = BaseType("x", data=np.arange(30))
    example["y"] = BaseType("y", data=np.arange(50))
    return example


def test_DatasetType_nbytes(gridtype_example):
    dataset = DatasetType("dataset")
    dataset["example"] = gridtype_example
    dataset["variable"] = BaseType("variable", np.arange(100, dtype=np.complex128))

    a_nb = dataset["example"].a.nbytes
    x_nb = dataset["example"].x.nbytes
    y_nb = dataset["example"].y.nbytes
    v_nb = dataset["variable"].nbytes

    # aggregate size of arrays only
    Nbytes = a_nb + x_nb + y_nb + v_nb

    assert dataset.nbytes == Nbytes


def test_GridType_repr(gridtype_example):
    """Test ``__repr__`` of grids."""
    assert repr(gridtype_example) == "<GridType with array 'a' and maps 'x', 'y'>"


def test_GridType_dtype(gridtype_example):
    """Test ``dtype`` of grids."""
    assert gridtype_example.dtype == np.dtype("l")


def test_GridType_shape(gridtype_example):
    """Test ``shape`` of grids."""
    assert gridtype_example.shape == (30, 50)


def test_GridType_size(gridtype_example):
    """Test ``size`` of grids."""
    assert gridtype_example.size == 30 * 50


def test_GridType_ndim(gridtype_example):
    """Test ``ndim`` of grids."""
    assert gridtype_example.ndim == 2


def test_GridType_len(gridtype_example):
    """Test ``__len__`` of grids."""
    assert len(gridtype_example) == 3


def test_GridType_getitem(gridtype_example):
    """Test item retrieval.

    As with sequences, this might return a child or a new grid, depending
    on the key type.

    """
    # a string should return the corresponding child
    assert isinstance(gridtype_example["a"], BaseType)

    # otherwise it should return a new grid with a subset of the data
    subset = gridtype_example[20:22, 40:43]
    assert subset["a"].shape == (2, 3)
    assert subset["x"].shape == (2,)
    assert subset["y"].shape == (3,)

    assert gridtype_example is not subset

    # pick more than one child:
    np.testing.assert_equal(subset["a", "x"]["a"].data, subset["a"].data)
    np.testing.assert_equal(subset["a", "x"]["x"].data, subset["x"].data)


def test_GridType_getitem_not_tuple(gridtype_example):
    """Test that method works with non-tuple slices."""
    subset = gridtype_example[20:22]
    assert subset["a"].shape == (2, 50)
    assert subset["x"].shape == (2,)
    assert subset["y"].shape == (50,)


def test_GridType_array(gridtype_example):
    """Test ``array`` property."""
    assert gridtype_example.array is gridtype_example["a"]


def test_GridType_array2(gridtype_example):
    """Test __array__ method."""
    np.testing.assert_array_equal(
        np.array(gridtype_example), gridtype_example["a"].data
    )


def test_GridType_maps(gridtype_example):
    """Test ``maps`` property."""
    assert list(gridtype_example.maps.items()) == [
        ("x", gridtype_example["x"]),
        ("y", gridtype_example["y"]),
    ]


def test_GridType_dimensions(gridtype_example):
    """Test ``dimensions`` property."""
    assert gridtype_example.dimensions == ("x", "y")



"""Test the DRM response with DAP4 protocol."""

import os
import unittest

from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__
from pydap.responses.dmr import dmr
from pydap.tests.datasets import SimpleGroup


def load_dmr_file(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    with open(abs_path, "r") as dmr_file:
        text = dmr_file.read()
    return text


class TestDMRResponseGroup(unittest.TestCase):
    """Test DRM response from sequences."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleGroup))
        self.res = app.get("/.dmr")

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_dispatcher(self):
        """Test the single dispatcher."""
        with self.assertRaises(StopIteration):
            dmr(None)

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "text/plain")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, "ascii")

    def test_headers(self):
        """Test the headers from the response."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "dmr"),
                    ("Content-type", "text/plain; charset=ascii"),
                    ("Access-Control-Allow-Origin", "*"),
                    (
                        "Access-Control-Allow-Headers",
                        "Origin, X-Requested-With, Content-Type",
                    ),
                    ("Content-Length", "1977"),
                ]
            ),
        )

    def test_body(self):
        """Test the generated DMR response"""
        dmr_text = load_dmr_file("data/dmrs/SimpleGroup.dmr")
        self.assertEqual(self.res.text, dmr_text)



"""
This is a simple example from the DODS Test Server.

    http://test.opendap.org:8080/dods/dts/D1

"""

import unittest

import numpy as np
from webob.request import Request

from pydap.client import open_url
from pydap.handlers.lib import BaseHandler
from pydap.tests.datasets import D1


class TestD1(unittest.TestCase):
    def setUp(self):
        # create WSGI app
        self.app = BaseHandler(D1)

    def test_dds(self):
        self.assertEqual(
            Request.blank("/.dds").get_response(self.app).text,
            """Dataset {
    Sequence {
        String instrument_id;
        String location;
        Float64 latitude;
        Float64 longitude;
    } Drifters;
} EOSDB%2EDBO;
""",
        )

    def test_ascii(self):
        resp = Request.blank("/.asc").get_response(self.app)
        content = resp.text
        self.assertEqual(
            content,
            """Dataset {
    Sequence {
        String instrument_id;
        String location;
        Float64 latitude;
        Float64 longitude;
    } Drifters;
} EOSDB%2EDBO;
---------------------------------------------
Drifters.instrument_id, Drifters.location, Drifters.latitude, Drifters.longitude
"This is a data test string (pass 1).", "This is a data test string (pass 0).", 1000, 999.95
"This is a data test string (pass 3).", "This is a data test string (pass 2).", 999.95, 999.55
"This is a data test string (pass 5).", "This is a data test string (pass 4).", 999.8, 998.75
"This is a data test string (pass 7).", "This is a data test string (pass 6).", 999.55, 997.55
"This is a data test string (pass 9).", "This is a data test string (pass 8).", 999.2, 995.95

""",
        )

    def test_data(self):
        dataset = open_url("http://localhost:8001/", application=self.app)
        data = list(dataset.Drifters.iterdata())
        self.assertEqual(data, D1.Drifters.data.tolist())

    def test_filtering(self):
        dataset = open_url("http://localhost:8001/", application=self.app)
        drifters = dataset.Drifters
        selection = np.rec.fromrecords(
            list(drifters[drifters.longitude < 999].iterdata()),
            names=list(drifters.keys()),
        )

        data = np.rec.fromrecords(
            D1.Drifters.data.tolist(), names=list(drifters.keys())
        )
        filtered = data[data["longitude"] < 999]

        np.testing.assert_array_equal(filtered, selection)

    def test_filtering_child(self):
        dataset = open_url("http://localhost:8001/", application=self.app)
        drifters = dataset.Drifters
        selection = np.array(list(drifters[drifters.longitude < 999]["location"]))

        data = np.rec.fromrecords(
            D1.Drifters.data.tolist(), names=list(drifters.keys())
        )
        filtered = data[data["longitude"] < 999]["location"]

        np.testing.assert_array_equal(filtered, selection)



"""Test parsing functions."""

import operator
import unittest

from pydap.exceptions import ConstraintExpressionError
from pydap.parsers import (
    SimpleParser,
    parse_ce,
    parse_hyperslab,
    parse_projection,
    parse_selection,
)
from pydap.tests.datasets import VerySimpleSequence


class TestParseProjection(unittest.TestCase):
    """Test projection parser.

    The projection is a comma separated list of variable ids (or names, if the
    client is using the shorthand notation), with optional hyperslabs and
    possibly function calls.

    """

    def test_basic(self):
        """Test a basic projection."""
        parsed = parse_projection("a,b,c")
        self.assertEqual(parsed, [[("a", ())], [("b", ())], [("c", ())]])

    def test_deep(self):
        """Test a projection with a deep id."""
        parsed = parse_projection("a.b.c,d")
        self.assertEqual(parsed, [[("a", ()), ("b", ()), ("c", ())], [("d", ())]])

    def test_hyperslab(self):
        """Test a projection with a hyperslab."""
        parsed = parse_projection("a,b.c[0:2:9]")
        self.assertEqual(parsed, [[("a", ())], [("b", ()), ("c", (slice(0, 10, 2),))]])

    def test_function_call(self):
        """Test a projection with a function call."""
        parsed = parse_projection("mean(a[0],1)")
        self.assertEqual(parsed, ["mean(a[0],1)"])

    def test_nested_function_call(self):
        """Test a projection with a function call."""
        parsed = parse_projection("mean(mean(a[0],1))")
        self.assertEqual(parsed, ["mean(mean(a[0],1))"])


class TestParseSelection(unittest.TestCase):
    """Test selection parser."""

    def test_simple(self):
        """Test a simple selection."""
        id1, op, id2 = parse_selection("sequence.byte>1", VerySimpleSequence)
        self.assertIs(id1, VerySimpleSequence.sequence.byte)
        self.assertIs(op, operator.gt)
        self.assertEqual(id2, 1)

    def test_inverted(self):
        """Test an inverted selection."""
        id1, op, id2 = parse_selection("1<sequence.byte", VerySimpleSequence)
        self.assertEqual(id1, 1)
        self.assertIs(op, operator.lt)
        self.assertIs(id2, VerySimpleSequence.sequence.byte)


class TestParseCe(unittest.TestCase):
    """Test the constraint expression parser."""

    def test_empty(self):
        """Test no constraint expression."""
        projection, selection = parse_ce("")
        self.assertEqual(projection, [])
        self.assertEqual(selection, [])

    def test_basic(self):
        """Test a basic constraint expression."""
        projection, selection = parse_ce("a,b&c>1")
        self.assertEqual(projection, [[("a", ())], [("b", ())]])
        self.assertEqual(selection, ["c>1"])

    def test_only_projection(self):
        """Test a constraint expression with only the projection."""
        projection, selection = parse_ce("a,b")
        self.assertEqual(projection, [[("a", ())], [("b", ())]])
        self.assertEqual(selection, [])

    def test_only_selection(self):
        """Test a constraint expression with only the selection."""
        projection, selection = parse_ce("c>1")
        self.assertEqual(projection, [])
        self.assertEqual(selection, ["c>1"])

    def test_complex(self):
        """Test a more complex constraint expression."""
        projection, selection = parse_ce("a,b[0:2:9],c&a>1&b<2")
        self.assertEqual(
            projection, [[("a", ())], [("b", (slice(0, 10, 2),))], [("c", ())]]
        )
        self.assertEqual(selection, ["a>1", "b<2"])

    def test_function(self):
        """Test a constraint expression with a function call."""
        projection, selection = parse_ce(
            "time&bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)"
        )
        self.assertEqual(projection, [[("time", ())]])
        self.assertEqual(
            selection, ["bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)"]
        )

    def test_function_no_selection(self):
        """Test a constraint expression with a function call."""
        projection, selection = parse_ce(
            "time,bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)"
        )
        self.assertEqual(
            projection,
            [[("time", ())], "bounds(0,360,-90,90,0,500,00Z01JAN1970,00Z04JAN1970)"],
        )
        self.assertEqual(selection, [])

    def test_dap4_ce(self):
        protocol = "dap4"
        projection, sel = parse_ce("dap4.ce=/lon[100:1:199]", protocol)
        self.assertEqual(projection, [[("/lon", (slice(100, 200, 1),))]])
        self.assertEqual(sel, [])

    def test_inconsistent_CE_protocol(self):
        query1 = "dap4.ce=/lon[100:1:199]"
        query2 = "lon[100:1:199]"
        with self.assertRaises(ConstraintExpressionError):
            parse_ce(query1, "dap2")
            parse_ce(query2, "dap4")


class TestParseHyperslab(unittest.TestCase):
    """Test hyperslab parser."""

    def test_point(self):
        """Test a single value selection."""
        self.assertEqual(parse_hyperslab("[0]"), (slice(0, 1, 1),))

    def test_start_stop(self):
        """Test start and stop, default step."""
        self.assertEqual(parse_hyperslab("[0:1]"), (slice(0, 2, 1),))

    def test_start_step_stop(self):
        """Test start, step and stop."""
        self.assertEqual(parse_hyperslab("[0:2:9]"), (slice(0, 10, 2),))

    def test_invalid(self):
        """Test invalid hyperslab."""
        with self.assertRaises(ConstraintExpressionError):
            parse_hyperslab("[0:2:9:1]")

    def test_ndimensionsal(self):
        """Test n-dimensional slices."""
        self.assertEqual(
            parse_hyperslab("[0:2:9][0][0:99]"),
            (slice(0, 10, 2), slice(0, 1, 1), slice(0, 100, 1)),
        )


class TestSimpleParser(unittest.TestCase):
    """Test the base class for the DAS/DDS parsers."""

    def test_peek_existing(self):
        """Test the peek method when the token exists."""
        parser = SimpleParser("Hello, World!")
        self.assertEqual(parser.peek(r"\w+"), "Hello")

    def test_peek_missing(self):
        """Test the peek method when the token does not exist."""
        parser = SimpleParser("Hello, World!")
        self.assertEqual(parser.peek(r"\d+"), "")

    def test_consume_existing(self):
        """Test the consume method when the token exists."""
        parser = SimpleParser("Hello, World!")
        self.assertEqual(parser.consume(r"\w+"), "Hello")
        self.assertEqual(parser.consume(", "), ", ")
        self.assertEqual(parser.consume(r"\w+"), "World")
        self.assertEqual(parser.consume("!"), "!")

    def test_consume_missing(self):
        """Test the consume method when the token does not exist."""
        parser = SimpleParser("Hello, World!")
        with self.assertRaises(Exception):
            parser.consume(r"\d+")



#
# Test loading a URL from the EarthData Cloud (EDC) OPeNDAP server using the
# new (8/18/22) PyDAP DAP4 support. This dataset will not load using DAP2.
#

import configparser
import logging
import time

import requests

import pydap.client

logging.basicConfig(
    filename="dap4_access_test.log",
    filemode="w",
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)s %(name)s : %(message)s",
)

logger = logging.getLogger(__name__)


class SessionEarthData(requests.Session):
    AUTH_HOST = "urs.earthdata.nasa.gov"

    def __init__(self, username, password):
        super().__init__()
        self.auth = (username, password)

    def rebuild_auth(self, prepared_request, response):
        headers = prepared_request.headers
        url = prepared_request.url
        if "Authorization" in headers:
            original_parsed = requests.utils.urlparse(response.request.url)
            redirect_parsed = requests.utils.urlparse(url)
            if (
                (original_parsed.hostname != redirect_parsed.hostname)
                and redirect_parsed.hostname != self.AUTH_HOST
                and original_parsed.hostname != self.AUTH_HOST
            ):
                del headers["Authorization"]
        return


def main():
    config = configparser.ConfigParser()
    config.read("user.config")
    username = config["user"]["user"]
    password = config["user"]["pwd"]
    session = SessionEarthData(username=username, password=password)

    dap4_schema = "dap4"

    host = "opendap.earthdata.nasa.gov"
    path = "/collections/C1996881146-POCLOUD/granules/"
    dataset = "20220531090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1"

    dap4_url = f"{dap4_schema}://{host}{path}{dataset}"

    # open the dap4 URL
    print(f"Open this URL: {dap4_url}")
    i = 0
    while True:
        start = time.time()
        pydap_ds = pydap.client.open_url(dap4_url, session=session)

        print("The attributes:")
        print(pydap_ds["sea_ice_fraction"].attributes)
        i += 1
        print(f"Call {i}")
        variable = pydap_ds["sea_ice_fraction"][0, 1700:1799:10, 1800:1900:10]
        print(f"Time for request: {time.time() - start}")
        print('A subset of the "sea_ice_fraction" variable')
        print(variable.data)


if __name__ == "__main__":
    main()



"""
This test module uses a simple development server
to test http-specific options like timeout
or caching.

It is based on simple data but with more handlers
it could work with more data formats.
"""

import ssl
import sys

import numpy as np
import pytest
import requests

from pydap.client import open_url
from pydap.handlers.lib import BaseHandler
from pydap.model import BaseType, DatasetType, SequenceType
from pydap.server.devel_ssl import LocalTestServerSSL

server = pytest.mark.server


@pytest.fixture
def sequence_type_data():
    """
    Simple sequence test data
    """
    data = [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]
    dtype = [("index", "<i4"), ("temperature", "<f8"), ("station", "S40")]
    seq = SequenceType("sequence")
    seq["index"] = BaseType("index")
    seq["temperature"] = BaseType("temperature")
    seq["station"] = BaseType("station")
    seq.data = np.array(data, dtype=dtype)
    return seq


@server
def test_open(sequence_type_data):
    """Test that LocalTestServerSSL works properly"""
    TestDataset = DatasetType("Test")
    TestDataset["sequence"] = sequence_type_data
    with LocalTestServerSSL(BaseHandler(TestDataset)) as server:
        dataset = open_url(server.url)
        seq = dataset["sequence"]
        retrieved_data = [line for line in seq]

    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=sequence_type_data.data.dtype),
        np.array(sequence_type_data.data[:], dtype=sequence_type_data.data.dtype),
    )


@pytest.mark.filterwarnings("ignore::urllib3.exceptions.InsecureRequestWarning")
@server
def test_verify_open_url(sequence_type_data):
    """Test that open_url raises the correct SSLError"""
    # warnings.simplefilter("always")

    TestDataset = DatasetType("Test")
    TestDataset["sequence"] = sequence_type_data
    TestDataset["byte"] = BaseType("byte", 0)
    application = BaseHandler(TestDataset)
    with LocalTestServerSSL(application, ssl_context="adhoc") as server:
        try:
            open_url(server.url, verify=False, session=requests.Session())
        except (ssl.SSLError, requests.exceptions.SSLError):
            pytest.fail("SSLError should not be raised.")

        with pytest.raises(requests.exceptions.SSLError):
            open_url(server.url, session=requests.Session())

        if not (sys.version_info >= (3, 0) and sys.version_info < (3, 4, 4)):
            # verify is disabled by default for python 3 before 3.4.4:
            with pytest.raises(requests.exceptions.SSLError):
                open_url(server.url)
